"""Auto-healing response engine for anomaly mitigation."""

from __future__ import annotations

from dataclasses import asdict, dataclass, field
import copy
import json
import threading
import time
from typing import Any, Optional
from urllib import request as urlrequest

from orchesis.message_chain import validate_tool_chain

_SUSPICIOUS_TEXT_MARKERS = (
    "ignore previous instructions",
    "override system",
    "disable security",
    "enter your password",
    "sk-",
    "-----begin private key-----",
)

DIAGNOSIS_RULES: dict[str, list[str]] = {
    "anomaly_high": ["strip_content", "rate_limit"],
    "anomaly_critical": ["strip_content", "rate_limit", "escalate"],
    "entropy_drop": ["reset_context"],
    "entropy_spike": ["strip_content"],
    "entropy_zero": ["reset_context", "rate_limit"],
    "tool_chain_loop": ["reset_context"],
    "escalation_chain": ["reset_context"],
    "ping_pong": ["reset_context", "rate_limit"],
    "request_template_repeat": ["rate_limit"],
    "drift_injection": ["strip_content", "inject_guardrail"],
    "drift_model_switch": ["inject_guardrail"],
    "drift_persona": ["inject_guardrail", "escalate"],
    "FM-1.3": ["inject_guardrail", "escalate"],
    "FM-1.4": ["rate_limit", "strip_content"],
    "FM-1.5": ["strip_content"],
    "FM-2.3": ["reset_context", "retry_model"],
    "FM-2.6": ["rate_limit", "escalate"],
    "FM-3.1": ["strip_content", "escalate"],
    "OE-1": ["inject_guardrail"],
    "OE-6": ["inject_guardrail", "escalate"],
    "upstream_429": ["retry_model"],
    "upstream_500": ["retry_model"],
    "upstream_context_length": ["reset_context", "retry_model"],
    "budget_exceeded": ["retry_model"],
    "context_overflow": ["reset_context", "retry_model"],
    "loop_detected": ["reset_context"],
    "cron_accumulation": ["reset_context"],
    "retry_storm": ["rate_limit"],
    "heartbeat_storm": ["rate_limit"],
    "injection_detected": ["strip_content"],
    "credential_leakage": ["strip_content"],
    "output_manipulation": ["strip_content", "escalate"],
    "privilege_escalation": ["inject_guardrail", "escalate"],
    "compliance_drift": ["inject_guardrail", "escalate"],
    "tool_abuse": ["rate_limit", "strip_content"],
    "cascading_failure": ["rate_limit", "escalate"],
}

_DEFAULT_ACTION_ORDER = (
    "strip_content",
    "reset_context",
    "retry_model",
    "inject_guardrail",
    "rate_limit",
    "escalate",
)


@dataclass
class HealingAction:
    """A single healing action to apply."""

    action_type: str
    reason: str
    parameters: dict[str, Any] = field(default_factory=dict)
    confidence: float = 0.5
    source_detection: str = ""


@dataclass
class HealingResult:
    """Result of applying healing actions."""

    actions_taken: list[HealingAction] = field(default_factory=list)
    original_issue: str = ""
    resolution: str = "failed"
    messages_modified: bool = False
    model_switched: bool = False
    context_reset: bool = False
    latency_added_ms: float = 0.0


class AgentRateLimiter:
    """Per-agent rate limiting."""

    def __init__(self) -> None:
        self._limits: dict[str, dict[str, float]] = {}
        self._lock = threading.Lock()

    def set_limit(self, agent_id: str, max_requests: int, window_seconds: int) -> None:
        key = str(agent_id or "unknown")
        with self._lock:
            self._limits[key] = {
                "window_start": time.monotonic(),
                "count": 0.0,
                "max_requests": float(max(1, int(max_requests))),
                "window_seconds": float(max(1, int(window_seconds))),
            }

    def check(self, agent_id: str) -> bool:
        key = str(agent_id or "unknown")
        now = time.monotonic()
        with self._lock:
            item = self._limits.get(key)
            if item is None:
                return True
            if (now - float(item["window_start"])) >= float(item["window_seconds"]):
                item["window_start"] = now
                item["count"] = 0.0
            item["count"] = float(item["count"]) + 1.0
            return item["count"] <= float(item["max_requests"])

    def remove_limit(self, agent_id: str) -> None:
        key = str(agent_id or "unknown")
        with self._lock:
            self._limits.pop(key, None)

    def get_limited_agents(self) -> list[str]:
        with self._lock:
            return sorted(self._limits.keys())

    def reset(self) -> None:
        with self._lock:
            self._limits = {}


class AutoHealer:
    """Automated response engine for detected anomalies."""

    def __init__(self, config: Optional[dict] = None):
        cfg = config if isinstance(config, dict) else {}
        self.enabled = bool(cfg.get("enabled", True))
        self.mode = str(cfg.get("mode", "active") or "active").strip().lower()
        if self.mode not in {"active", "monitor"}:
            self.mode = "active"
        self.max_retries = max(1, int(cfg.get("max_retries", 2)))
        self.max_healing_time_ms = max(1, int(cfg.get("max_healing_time_ms", 3000)))
        actions_cfg = cfg.get("actions") if isinstance(cfg.get("actions"), dict) else {}
        self._actions_cfg = actions_cfg
        self._action_confidence: dict[str, float] = {name: 0.7 for name in _DEFAULT_ACTION_ORDER}
        self._action_confidence["pass"] = 1.0
        self.rate_limiter = AgentRateLimiter()
        self._lock = threading.Lock()
        self._agent_history: dict[str, list[dict[str, Any]]] = {}
        self._stats = {
            "total_diagnoses": 0,
            "total_actions_taken": 0,
            "actions_by_type": {},
            "healed_or_mitigated": 0,
            "total_healing_time_ms": 0.0,
            "healing_attempts": 0,
            "escalations": 0,
        }

    def _action_enabled(self, action_type: str) -> bool:
        section = self._actions_cfg.get(action_type)
        if isinstance(section, dict):
            return bool(section.get("enabled", True))
        return True

    def _action_cfg(self, action_type: str) -> dict[str, Any]:
        section = self._actions_cfg.get(action_type)
        return section if isinstance(section, dict) else {}

    def _collect_signals(self, detection_result: Any, mast_findings: Any, request_data: dict[str, Any]) -> list[str]:
        signals: list[str] = []
        det = detection_result
        if det is not None:
            try:
                if bool(getattr(det, "is_anomalous", False)):
                    level = str(getattr(det, "risk_level", "low"))
                    if level in {"high", "critical"}:
                        signals.append(f"anomaly_{level}")
                drift = str(getattr(det, "drift_type", "") or "")
                if drift and drift != "normal":
                    signals.append(f"drift_{drift}")
                if bool(getattr(det, "entropy_anomalous", False)):
                    score = float(getattr(det, "entropy_score", 0.0) or 0.0)
                    if score <= 0.01:
                        signals.append("entropy_zero")
                    elif score < 25.0:
                        signals.append("entropy_drop")
                    elif score >= 80.0:
                        signals.append("entropy_spike")
                patterns = getattr(det, "patterns_found", []) or []
                for pat in patterns:
                    ptype = str(getattr(pat, "pattern_type", "") or "")
                    if ptype == "request_template":
                        signals.append("request_template_repeat")
                    elif ptype:
                        signals.append(ptype)
            except Exception:
                pass
        if isinstance(mast_findings, list):
            for finding in mast_findings:
                mode = str(getattr(finding, "failure_mode", "") or "")
                if mode:
                    signals.append(mode)
        if isinstance(request_data, dict):
            issue = str(request_data.get("healing_issue", "") or "")
            if issue:
                signals.append(issue)
            status_code = int(request_data.get("upstream_status", 0) or 0)
            if status_code == 429:
                signals.append("upstream_429")
            elif status_code >= 500:
                signals.append("upstream_500")
            if bool(request_data.get("context_overflow")):
                signals.append("context_overflow")
        return list(dict.fromkeys(signals))

    def _candidate_actions_for_signal(self, signal: str) -> list[str]:
        return list(DIAGNOSIS_RULES.get(signal, []))

    def diagnose(
        self,
        detection_result: Any = None,
        mast_findings: Any = None,
        request_data: Any = None,
        context: Any = None,
    ) -> list[HealingAction]:
        req = request_data if isinstance(request_data, dict) else {}
        _ = context
        signals = self._collect_signals(detection_result, mast_findings, req)
        with self._lock:
            self._stats["total_diagnoses"] += 1
        if not signals:
            return []
        actions: list[HealingAction] = []
        seen: set[str] = set()
        for signal in signals:
            for action_type in self._candidate_actions_for_signal(signal):
                if action_type in seen:
                    continue
                if not self._action_enabled(action_type):
                    continue
                if action_type == "retry_model":
                    fallback_models = self._action_cfg("retry_model").get("fallback_models", [])
                    if not isinstance(fallback_models, list) or not fallback_models:
                        continue
                conf = float(self._action_confidence.get(action_type, 0.6))
                reason = f"trigger={signal}"
                params: dict[str, Any] = {}
                if action_type == "retry_model":
                    params["fallback_models"] = list(
                        self._action_cfg("retry_model").get("fallback_models", [])
                    )
                elif action_type == "reset_context":
                    cfg = self._action_cfg("reset_context")
                    params = {
                        "preserve_system": bool(cfg.get("preserve_system", True)),
                        "preserve_last_n": int(cfg.get("preserve_last_n", 3)),
                    }
                elif action_type == "rate_limit":
                    cfg = self._action_cfg("rate_limit")
                    params = {
                        "window_seconds": int(cfg.get("window_seconds", 60)),
                        "max_requests": int(cfg.get("max_requests", 10)),
                    }
                elif action_type == "inject_guardrail":
                    cfg = self._action_cfg("inject_guardrail")
                    params = {
                        "guardrail_prefix": str(
                            cfg.get(
                                "guardrail_prefix",
                                "IMPORTANT: Do not attempt to access tools or resources outside your approved scope.",
                            )
                        )
                    }
                elif action_type == "escalate":
                    cfg = self._action_cfg("escalate")
                    params = {"webhook_url": cfg.get("webhook_url")}
                actions.append(
                    HealingAction(
                        action_type=action_type,
                        reason=reason,
                        parameters=params,
                        confidence=max(0.0, min(1.0, conf)),
                        source_detection=signal,
                    )
                )
                seen.add(action_type)
        if not actions:
            return []
        order_idx = {name: idx for idx, name in enumerate(_DEFAULT_ACTION_ORDER)}
        actions.sort(
            key=lambda item: (
                -float(item.confidence),
                order_idx.get(item.action_type, 99),
            )
        )
        return actions[:3]

    @staticmethod
    def _copy_request(request_data: dict[str, Any]) -> dict[str, Any]:
        return copy.deepcopy(request_data if isinstance(request_data, dict) else {})

    @staticmethod
    def _sanitize_text(text: str) -> str:
        value = str(text or "")
        lowered = value.lower()
        if any(marker in lowered for marker in _SUSPICIOUS_TEXT_MARKERS):
            return "[stripped suspicious content]"
        return value

    def _apply_retry_model(self, data: dict[str, Any], action: HealingAction, result: HealingResult) -> None:
        current_model = str(data.get("model", "") or "")
        fallbacks = action.parameters.get("fallback_models", [])
        if not isinstance(fallbacks, list):
            return
        for candidate in fallbacks:
            model = str(candidate or "")
            if model and model != current_model:
                data["model"] = model
                result.model_switched = True
                return

    def _apply_reset_context(self, data: dict[str, Any], action: HealingAction, result: HealingResult) -> None:
        messages = data.get("messages")
        if not isinstance(messages, list):
            return
        preserve_system = bool(action.parameters.get("preserve_system", True))
        preserve_last_n = max(0, int(action.parameters.get("preserve_last_n", 3)))
        tail = messages[-preserve_last_n:] if preserve_last_n > 0 else []
        kept: list[dict[str, Any]] = []
        if preserve_system:
            for msg in messages:
                if isinstance(msg, dict) and str(msg.get("role", "")).lower() == "system":
                    kept.append(dict(msg))
        for msg in tail:
            if isinstance(msg, dict):
                kept.append(dict(msg))
        data["messages"] = validate_tool_chain(kept)
        result.messages_modified = True
        result.context_reset = True

    def _apply_strip_content(self, data: dict[str, Any], result: HealingResult) -> None:
        messages = data.get("messages")
        if not isinstance(messages, list):
            return
        new_messages: list[dict[str, Any]] = []
        changed = False
        for msg in messages:
            if not isinstance(msg, dict):
                continue
            copied = dict(msg)
            content = copied.get("content")
            if isinstance(content, str):
                safe = self._sanitize_text(content)
                if safe != content:
                    changed = True
                copied["content"] = safe
            new_messages.append(copied)
        if changed:
            result.messages_modified = True
            data["messages"] = validate_tool_chain(new_messages)

    def _apply_rate_limit(self, agent_id: str, action: HealingAction) -> None:
        max_requests = int(action.parameters.get("max_requests", 10))
        window_seconds = int(action.parameters.get("window_seconds", 60))
        self.rate_limiter.set_limit(agent_id, max_requests=max_requests, window_seconds=window_seconds)

    def _apply_inject_guardrail(self, data: dict[str, Any], action: HealingAction, result: HealingResult) -> None:
        messages = data.get("messages")
        if not isinstance(messages, list):
            return
        guardrail = str(action.parameters.get("guardrail_prefix", "") or "").strip()
        if not guardrail:
            return
        out: list[dict[str, Any]] = []
        injected = False
        for msg in messages:
            if not isinstance(msg, dict):
                continue
            copied = dict(msg)
            role = str(copied.get("role", "")).lower()
            if not injected and role == "system":
                original = str(copied.get("content", "") or "")
                if not original.startswith(guardrail):
                    copied["content"] = f"{guardrail}\n\n{original}".strip()
                    result.messages_modified = True
                injected = True
            out.append(copied)
        if not injected:
            out.insert(0, {"role": "system", "content": guardrail})
            result.messages_modified = True
        data["messages"] = validate_tool_chain(out)

    @staticmethod
    def _async_webhook(url: str, payload: dict[str, Any]) -> None:
        def _worker() -> None:
            try:
                req = urlrequest.Request(
                    url=url,
                    data=json.dumps(payload, ensure_ascii=False).encode("utf-8"),
                    headers={"Content-Type": "application/json"},
                    method="POST",
                )
                with urlrequest.urlopen(req, timeout=2.0):  # noqa: S310
                    pass
            except Exception:
                return

        t = threading.Thread(target=_worker, daemon=True)
        t.start()

    def apply(self, actions: list[HealingAction], request_data: dict, agent_id: str) -> tuple[dict, HealingResult]:
        started = time.perf_counter()
        data = self._copy_request(request_data)
        result = HealingResult(
            actions_taken=[],
            original_issue=",".join([a.source_detection for a in actions]) if actions else "",
            resolution="failed",
        )
        if not actions:
            result.resolution = "failed"
            return data, result
        attempts = 0
        for action in actions:
            if attempts >= self.max_retries:
                break
            elapsed_ms = (time.perf_counter() - started) * 1000.0
            if elapsed_ms > float(self.max_healing_time_ms):
                break
            attempts += 1
            a_type = action.action_type
            if a_type == "pass":
                result.actions_taken.append(action)
                continue
            if a_type == "retry_model":
                self._apply_retry_model(data, action, result)
                result.actions_taken.append(action)
            elif a_type == "reset_context":
                self._apply_reset_context(data, action, result)
                result.actions_taken.append(action)
            elif a_type == "strip_content":
                self._apply_strip_content(data, result)
                result.actions_taken.append(action)
            elif a_type == "rate_limit":
                self._apply_rate_limit(str(agent_id or "unknown"), action)
                result.actions_taken.append(action)
            elif a_type == "inject_guardrail":
                self._apply_inject_guardrail(data, action, result)
                result.actions_taken.append(action)
            elif a_type == "escalate":
                url = str(action.parameters.get("webhook_url", "") or "")
                payload = {
                    "agent_id": str(agent_id or "unknown"),
                    "reason": action.reason,
                    "source_detection": action.source_detection,
                    "timestamp": time.time(),
                }
                if url:
                    self._async_webhook(url, payload)
                with self._lock:
                    self._stats["escalations"] += 1
                result.actions_taken.append(action)
            with self._lock:
                self._stats["total_actions_taken"] += 1
                actions_map = self._stats["actions_by_type"]
                actions_map[a_type] = int(actions_map.get(a_type, 0)) + 1
        if result.messages_modified and isinstance(data.get("messages"), list):
            data["messages"] = validate_tool_chain(data["messages"])
        result.resolution = "healed" if (result.messages_modified or result.model_switched or result.context_reset) else "mitigated"
        result.latency_added_ms = round((time.perf_counter() - started) * 1000.0, 3)
        with self._lock:
            self._stats["healing_attempts"] += 1
            self._stats["total_healing_time_ms"] += result.latency_added_ms
            if result.resolution in {"healed", "mitigated"}:
                self._stats["healed_or_mitigated"] += 1
            self._agent_history.setdefault(str(agent_id or "unknown"), []).append(
                {
                    "timestamp": time.time(),
                    "result": asdict(result),
                    "actions": [asdict(a) for a in result.actions_taken],
                }
            )
            if len(self._agent_history[str(agent_id or "unknown")]) > 100:
                self._agent_history[str(agent_id or "unknown")] = self._agent_history[str(agent_id or "unknown")][-100:]
        return data, result

    def heal(
        self,
        detection_result: Any = None,
        mast_findings: Any = None,
        request_data: Any = None,
        agent_id: str | None = None,
        context: Any = None,
    ) -> tuple[dict, HealingResult]:
        req = self._copy_request(request_data if isinstance(request_data, dict) else {})
        actions = self.diagnose(
            detection_result=detection_result,
            mast_findings=mast_findings,
            request_data=req,
            context=context,
        )
        if self.mode == "monitor":
            monitor_action = HealingAction(
                action_type="pass",
                reason="monitor mode",
                parameters={},
                confidence=1.0,
                source_detection="monitor",
            )
            res = HealingResult(
                actions_taken=[monitor_action] if actions else [],
                original_issue=",".join([a.source_detection for a in actions]) if actions else "",
                resolution="mitigated" if actions else "failed",
                messages_modified=False,
                model_switched=False,
                context_reset=False,
                latency_added_ms=0.0,
            )
            with self._lock:
                self._agent_history.setdefault(str(agent_id or "unknown"), []).append(
                    {
                        "timestamp": time.time(),
                        "result": asdict(res),
                        "actions": [asdict(a) for a in res.actions_taken],
                    }
                )
            return req, res
        if not actions:
            return req, HealingResult(actions_taken=[], original_issue="", resolution="failed", latency_added_ms=0.0)
        return self.apply(actions=actions, request_data=req, agent_id=str(agent_id or "unknown"))

    def verify_healing(self, agent_id: str, pre_healing_score: float, post_healing_score: float) -> bool:
        improved = float(post_healing_score) <= float(pre_healing_score)
        key = str(agent_id or "unknown")
        with self._lock:
            history = self._agent_history.get(key, [])
            if history:
                last = history[-1]
                actions = last.get("actions", [])
                for item in actions:
                    action_type = str(item.get("action_type", ""))
                    if not action_type or action_type == "pass":
                        continue
                    current = float(self._action_confidence.get(action_type, 0.6))
                    if improved:
                        current = min(0.98, current + 0.03)
                    else:
                        current = max(0.10, current - 0.05)
                    self._action_confidence[action_type] = current
                last["verified"] = True
                last["improved"] = improved
                last["pre_score"] = float(pre_healing_score)
                last["post_score"] = float(post_healing_score)
        return improved

    def get_stats(self) -> dict[str, Any]:
        with self._lock:
            attempts = int(self._stats["healing_attempts"])
            avg_ms = (float(self._stats["total_healing_time_ms"]) / float(attempts)) if attempts > 0 else 0.0
            success_rate = (float(self._stats["healed_or_mitigated"]) / float(attempts)) if attempts > 0 else 0.0
            return {
                "enabled": self.enabled,
                "mode": self.mode,
                "total_diagnoses": int(self._stats["total_diagnoses"]),
                "total_actions_taken": int(self._stats["total_actions_taken"]),
                "actions_by_type": dict(self._stats["actions_by_type"]),
                "success_rate": round(success_rate, 4),
                "avg_healing_time_ms": round(avg_ms, 3),
                "escalations": int(self._stats["escalations"]),
                "currently_rate_limited": self.rate_limiter.get_limited_agents(),
                "confidence_by_action": dict(self._action_confidence),
            }

    def get_agent_healing_history(self, agent_id: str) -> list[dict]:
        key = str(agent_id or "unknown")
        with self._lock:
            return copy.deepcopy(self._agent_history.get(key, []))

    def reset(self, agent_id: str | None = None) -> None:
        if agent_id:
            key = str(agent_id or "unknown")
            with self._lock:
                self._agent_history.pop(key, None)
            self.rate_limiter.remove_limit(key)
            return
        with self._lock:
            self._agent_history = {}
            self._stats = {
                "total_diagnoses": 0,
                "total_actions_taken": 0,
                "actions_by_type": {},
                "healed_or_mitigated": 0,
                "total_healing_time_ms": 0.0,
                "healing_attempts": 0,
                "escalations": 0,
            }
            self._action_confidence = {name: 0.7 for name in _DEFAULT_ACTION_ORDER}
            self._action_confidence["pass"] = 1.0
        self.rate_limiter.reset()



"""Structured telemetry export for research and analysis.

Captures every request through the proxy as a structured record.
Exports to JSONL (one JSON object per line) and CSV.
Designed for offline analysis - zero impact on request path.
"""

from __future__ import annotations

import csv
import json
import threading
import time
from collections import deque
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any


@dataclass
class TelemetryRecord:
    """Single request telemetry record."""

    timestamp: float = 0.0
    request_id: str = ""
    session_id: str = ""
    agent_id: str = ""
    model_requested: str = ""
    model_used: str = ""

    total_ms: float = 0.0
    upstream_ms: float = 0.0
    proxy_overhead_ms: float = 0.0

    input_tokens: int = 0
    output_tokens: int = 0
    cost_usd: float = 0.0

    threat_matches: list[str] = field(default_factory=list)
    threat_categories: list[str] = field(default_factory=list)
    threat_max_severity: str = ""
    blocked: bool = False
    block_reason: str = ""

    cache_hit: bool = False
    cache_type: str = "miss"

    loop_detected: bool = False
    loop_count: int = 0
    content_hash_blocked: bool = False
    heartbeat_detected: bool = False

    session_risk_score: float = 0.0
    session_risk_level: str = "observe"

    turn_number: int = 0
    tool_calls_count: int = 0
    has_tool_results: bool = False
    is_streaming: bool = False

    failure_mode: str = ""

    budget_remaining_usd: float = 0.0
    spend_rate_5min_usd: float = 0.0
    budget_blocked: bool = False

    was_cascaded: bool = False
    cascade_reason: str = ""

    status_code: int = 0
    error_type: str = ""


class TelemetryCollector:
    """Collect telemetry records in memory with bounded buffer."""

    def __init__(
        self,
        max_records: int = 100_000,
        auto_export_path: str | None = None,
        auto_export_interval: float = 300.0,
        enabled: bool = True,
    ) -> None:
        self._enabled = bool(enabled)
        self._max_records = max(1000, int(max_records))
        self._records: deque[TelemetryRecord] = deque(maxlen=self._max_records)
        self._lock = threading.Lock()
        self._stats = {"total_recorded": 0, "total_exported": 0, "dropped": 0}

        self._auto_export_path = auto_export_path
        self._auto_export_interval = float(max(1.0, auto_export_interval))
        self._export_thread: threading.Thread | None = None
        self._stop_event = threading.Event()

        if self._enabled and self._auto_export_path:
            self._start_auto_export()

    @property
    def enabled(self) -> bool:
        return self._enabled

    def record(self, rec: TelemetryRecord) -> None:
        if not self._enabled:
            return
        rec.timestamp = rec.timestamp or time.time()
        with self._lock:
            if len(self._records) >= self._max_records:
                self._stats["dropped"] += 1
            self._records.append(rec)
            self._stats["total_recorded"] += 1

    def export_jsonl(self, path: str) -> int:
        with self._lock:
            records = list(self._records)
        p = Path(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        count = 0
        with p.open("w", encoding="utf-8") as handle:
            for rec in records:
                handle.write(json.dumps(asdict(rec), default=str) + "\n")
                count += 1
        with self._lock:
            self._stats["total_exported"] += count
        return count

    def export_csv(self, path: str) -> int:
        with self._lock:
            records = list(self._records)
        if not records:
            return 0
        p = Path(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        fieldnames = list(asdict(records[0]).keys())
        count = 0
        with p.open("w", encoding="utf-8", newline="") as handle:
            writer = csv.DictWriter(handle, fieldnames=fieldnames)
            writer.writeheader()
            for rec in records:
                row = asdict(rec)
                for key, value in row.items():
                    if isinstance(value, list):
                        row[key] = ";".join(str(item) for item in value)
                writer.writerow(row)
                count += 1
        with self._lock:
            self._stats["total_exported"] += count
        return count

    def get_records(self, last_n: int = 0) -> list[dict[str, Any]]:
        with self._lock:
            if last_n > 0:
                items = list(self._records)[-int(last_n) :]
            else:
                items = list(self._records)
        return [asdict(item) for item in items]

    def clear(self) -> int:
        with self._lock:
            count = len(self._records)
            self._records.clear()
            return count

    @property
    def stats(self) -> dict[str, Any]:
        with self._lock:
            return {
                **self._stats,
                "buffered": len(self._records),
                "max_records": self._max_records,
                "enabled": self._enabled,
            }

    def _start_auto_export(self) -> None:
        def _loop() -> None:
            while not self._stop_event.wait(self._auto_export_interval):
                if self._auto_export_path and len(self._records) > 0:
                    try:
                        self.export_jsonl(self._auto_export_path)
                    except Exception:
                        pass

        self._export_thread = threading.Thread(target=_loop, daemon=True, name="telemetry-export")
        self._export_thread.start()

    def stop(self) -> None:
        self._stop_event.set()
        if self._export_thread is not None:
            self._export_thread.join(timeout=5.0)
        if self._auto_export_path and len(self._records) > 0:
            try:
                self.export_jsonl(self._auto_export_path)
            except Exception:
                pass


def build_record_from_context(ctx: Any) -> TelemetryRecord:
    rec = TelemetryRecord()
    pr = getattr(ctx, "proc_result", {}) if ctx else {}
    if not isinstance(pr, dict):
        pr = {}

    rec.request_id = str(pr.get("request_id", ""))
    rec.session_id = str(pr.get("session_id", ""))
    rec.agent_id = str(pr.get("agent_id", ""))
    rec.model_requested = str(pr.get("model", ""))
    rec.model_used = str(pr.get("model_used", pr.get("model", "")))

    rec.total_ms = float(pr.get("total_ms", 0.0) or 0.0)
    rec.upstream_ms = float(pr.get("upstream_ms", 0.0) or 0.0)
    rec.proxy_overhead_ms = max(0.0, rec.total_ms - rec.upstream_ms)

    rec.input_tokens = int(pr.get("input_tokens", 0) or 0)
    rec.output_tokens = int(pr.get("output_tokens", 0) or 0)
    rec.cost_usd = float(pr.get("cost_usd", 0.0) or 0.0)

    threat_matches = pr.get("threat_matches", [])
    if isinstance(threat_matches, list) and threat_matches:
        rec.threat_matches = [str(getattr(m, "name", m)) for m in threat_matches]
        rec.threat_categories = sorted(
            {str(getattr(m, "category", "")).strip() for m in threat_matches if str(getattr(m, "category", "")).strip()}
        )
        sev = [str(getattr(m, "severity", "")).lower() for m in threat_matches]
        order = {"": 0, "info": 1, "low": 2, "medium": 3, "high": 4, "critical": 5}
        rec.threat_max_severity = max(sev, key=lambda item: order.get(item, 0), default="")

    rec.blocked = bool(pr.get("blocked", False))
    rec.block_reason = str(pr.get("block_reason", ""))

    rec.cache_hit = bool(pr.get("cache_hit", False))
    rec.cache_type = str(pr.get("cache_type", "miss"))

    rec.loop_detected = bool(pr.get("loop_detected", False))
    rec.loop_count = int(pr.get("loop_count", 0) or 0)
    rec.content_hash_blocked = bool(pr.get("content_hash_blocked", False))
    rec.heartbeat_detected = bool(pr.get("heartbeat_detected", False))

    rec.session_risk_score = float(pr.get("session_risk_score", 0.0) or 0.0)
    rec.session_risk_level = str(pr.get("session_risk_level", "observe"))

    rec.turn_number = int(pr.get("turn_number", 0) or 0)
    rec.tool_calls_count = int(pr.get("tool_calls_count", 0) or 0)
    rec.has_tool_results = bool(pr.get("has_tool_results", False))
    rec.is_streaming = bool(pr.get("streaming", False))

    rec.failure_mode = _classify_failure_mode(pr)

    rec.budget_remaining_usd = float(pr.get("budget_remaining_usd", 0.0) or 0.0)
    rec.spend_rate_5min_usd = float(pr.get("spend_rate_5min", 0.0) or 0.0)
    rec.budget_blocked = bool(pr.get("budget_blocked", False))

    rec.was_cascaded = bool(pr.get("cascaded", False))
    rec.cascade_reason = str(pr.get("cascade_reason", ""))

    rec.status_code = int(pr.get("status_code", 200) or 200)
    rec.error_type = str(pr.get("error_type", ""))
    return rec


def _classify_failure_mode(pr: dict[str, Any]) -> str:
    if pr.get("loop_detected") or pr.get("content_hash_blocked"):
        return "FM-1.3-loop"
    if pr.get("context_erosion_detected"):
        return "FM-1.4-context-erosion"
    if pr.get("unaware_termination"):
        return "FM-1.5-unaware-termination"
    if pr.get("task_derailment_detected"):
        return "FM-2.3-task-derailment"
    if pr.get("termination_confusion"):
        return "FM-3.1-termination-confusion"
    if pr.get("hallucination_risk"):
        return "OE-1-hallucination-risk"

    input_tokens = int(pr.get("input_tokens", 0) or 0)
    output_tokens = int(pr.get("output_tokens", 0) or 0)
    if input_tokens > 0 and output_tokens > 0 and input_tokens > output_tokens * 10:
        return "OE-6-token-waste"

    if pr.get("heartbeat_detected"):
        return "OE-heartbeat-storm"
    if pr.get("budget_blocked"):
        return "OE-budget-exceeded"
    if pr.get("blocked") and pr.get("threat_matches"):
        return "SEC-threat-blocked"
    return ""



from __future__ import annotations

import io
import json
from concurrent.futures import ThreadPoolExecutor

from orchesis.mcp_server import MCPServer, MCP_VERSION


def _server_with_tools() -> MCPServer:
    return MCPServer(
        {
            "orchesis_echo": {
                "description": "Echo payload",
                "inputSchema": {
                    "type": "object",
                    "properties": {"text": {"type": "string"}},
                    "required": ["text"],
                },
                "handler": lambda params: f"echo:{params['text']}",
            },
            "orchesis_json": {
                "description": "JSON payload",
                "inputSchema": {"type": "object", "properties": {}},
                "handler": lambda _params: {"ok": True},
            },
            "orchesis_list": {
                "description": "List payload",
                "inputSchema": {"type": "object", "properties": {}},
                "handler": lambda _params: [{"type": "text", "text": "x"}],
            },
            "orchesis_fail": {
                "description": "Fail payload",
                "inputSchema": {"type": "object", "properties": {}},
                "handler": lambda _params: (_ for _ in ()).throw(RuntimeError("boom")),
            },
        }
    )


def test_initialize_handshake_returns_protocol_and_server_info() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "id": 1, "method": "initialize", "params": {}})
    assert isinstance(response, dict)
    assert response["result"]["protocolVersion"] == MCP_VERSION
    assert response["result"]["serverInfo"]["name"] == "orchesis"


def test_initialize_returns_tools_capability() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "id": 2, "method": "initialize", "params": {}})
    assert response is not None
    assert response["result"]["capabilities"]["tools"]["listChanged"] is False


def test_tools_list_returns_all_tools() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "id": 3, "method": "tools/list", "params": {}})
    assert response is not None
    names = {item["name"] for item in response["result"]["tools"]}
    assert {"orchesis_echo", "orchesis_json", "orchesis_list", "orchesis_fail"} == names


def test_tools_call_string_result_wrapped_as_text_content() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {
            "jsonrpc": "2.0",
            "id": 4,
            "method": "tools/call",
            "params": {"name": "orchesis_echo", "arguments": {"text": "hello"}},
        }
    )
    assert response is not None
    payload = response["result"]
    assert payload["isError"] is False
    assert payload["content"][0]["text"] == "echo:hello"


def test_tools_call_dict_result_serialized_as_json_text() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {"jsonrpc": "2.0", "id": 5, "method": "tools/call", "params": {"name": "orchesis_json", "arguments": {}}}
    )
    assert response is not None
    assert '"ok": true' in response["result"]["content"][0]["text"].lower()


def test_tools_call_list_result_returned_as_is() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {"jsonrpc": "2.0", "id": 6, "method": "tools/call", "params": {"name": "orchesis_list", "arguments": {}}}
    )
    assert response is not None
    assert response["result"]["content"] == [{"type": "text", "text": "x"}]


def test_unknown_tool_returns_invalid_params_error() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {"jsonrpc": "2.0", "id": 7, "method": "tools/call", "params": {"name": "missing", "arguments": {}}}
    )
    assert response is not None
    assert response["error"]["code"] == -32602


def test_invalid_json_payload_returns_parse_error() -> None:
    server = _server_with_tools()
    responses = server.handle_json_message("{not-json}")
    assert responses[0]["error"]["code"] == -32700


def test_unknown_method_returns_method_not_found() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "id": 8, "method": "unknown", "params": {}})
    assert response is not None
    assert response["error"]["code"] == -32601


def test_ping_returns_empty_result_object() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "id": 9, "method": "ping", "params": {}})
    assert response is not None
    assert response["result"] == {}


def test_initialized_notification_returns_no_response() -> None:
    server = _server_with_tools()
    response = server._handle_request({"jsonrpc": "2.0", "method": "initialized", "params": {}})
    assert response is None


def test_missing_required_params_returns_error() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {"jsonrpc": "2.0", "id": 10, "method": "tools/call", "params": {"name": "orchesis_echo", "arguments": {}}}
    )
    assert response is not None
    assert response["error"]["code"] == -32602
    assert "Missing required parameter: text" in response["error"]["message"]


def test_tool_exception_returns_is_error_true() -> None:
    server = _server_with_tools()
    response = server._handle_request(
        {"jsonrpc": "2.0", "id": 11, "method": "tools/call", "params": {"name": "orchesis_fail", "arguments": {}}}
    )
    assert response is not None
    assert response["result"]["isError"] is True
    assert "Error: boom" in response["result"]["content"][0]["text"]


def test_batch_request_returns_only_non_notification_responses() -> None:
    server = _server_with_tools()
    payload = json.dumps(
        [
            {"jsonrpc": "2.0", "id": 12, "method": "ping", "params": {}},
            {"jsonrpc": "2.0", "method": "initialized", "params": {}},
            {"jsonrpc": "2.0", "id": 13, "method": "tools/list", "params": {}},
        ]
    )
    responses = server.handle_json_message(payload)
    ids = sorted(item["id"] for item in responses)
    assert ids == [12, 13]


def test_run_writes_stdout_lines_for_each_response(monkeypatch) -> None:
    server = _server_with_tools()
    stdin = io.StringIO('{"jsonrpc":"2.0","id":14,"method":"ping","params":{}}\n')
    stdout = io.StringIO()
    monkeypatch.setattr("sys.stdin", stdin)
    monkeypatch.setattr("sys.stdout", stdout)
    server.run()
    output = stdout.getvalue().strip()
    assert output
    payload = json.loads(output)
    assert payload["id"] == 14


def test_concurrent_request_handling_returns_correct_ids() -> None:
    server = _server_with_tools()

    def call(i: int) -> int:
        response = server._handle_request({"jsonrpc": "2.0", "id": i, "method": "ping", "params": {}})
        assert response is not None
        return int(response["id"])

    with ThreadPoolExecutor(max_workers=8) as pool:
        ids = sorted(pool.map(call, range(100, 120)))
    assert ids == list(range(100, 120))




"""OTel-compatible span emission without SDK dependency."""

from __future__ import annotations

import json
import time
import uuid
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

from orchesis.telemetry import DecisionEvent, EventEmitter

if TYPE_CHECKING:
    from orchesis.otel_export import OTLPSpanExporter

# GenAI semantic conventions (OpenTelemetry spec)
GENAI_ATTRS = {
    "gen_ai.system": str,
    "gen_ai.operation.name": str,
    "gen_ai.request.model": str,
    "gen_ai.request.max_tokens": int,
    "gen_ai.request.temperature": float,
    "gen_ai.request.top_p": float,
    "gen_ai.response.model": str,
    "gen_ai.response.finish_reasons": str,
    "gen_ai.response.id": str,
    "gen_ai.usage.input_tokens": int,
    "gen_ai.usage.output_tokens": int,
    "gen_ai.usage.total_tokens": int,
    "orchesis.proxy.phase": str,
    "orchesis.decision": str,
    "orchesis.cost_usd": float,
    "orchesis.session_id": str,
    "orchesis.agent_id": str,
    "orchesis.cascade_level": str,
    "orchesis.cache_hit": bool,
    "orchesis.cache_type": str,
    "orchesis.threat_detected": bool,
    "orchesis.threat_ids": str,
    "orchesis.loop_detected": bool,
    "orchesis.experiment_id": str,
    "orchesis.variant_name": str,
    "orchesis.context_tokens_saved": int,
    "orchesis.circuit_state": str,
}


@dataclass
class SpanData:
    trace_id: str
    span_id: str
    parent_span_id: str | None
    operation: str
    start_time_ns: int
    end_time_ns: int
    status: str
    attributes: dict[str, str | int | float | bool]
    events: list[dict[str, Any]]


class TraceContext:
    """Lightweight trace context propagation."""

    def __init__(self, trace_id: str | None = None, parent_span_id: str | None = None):
        self.trace_id = trace_id or self._generate_trace_id()
        self.parent_span_id = parent_span_id

    @staticmethod
    def _generate_trace_id() -> str:
        return uuid.uuid4().hex

    @staticmethod
    def from_headers(headers: dict) -> "TraceContext":
        raw = headers.get("traceparent")
        if not isinstance(raw, str):
            return TraceContext()
        parts = raw.strip().split("-")
        if len(parts) != 4:
            return TraceContext()
        version, trace_id, parent_span_id, _flags = parts
        if len(version) != 2 or len(trace_id) != 32 or len(parent_span_id) != 16:
            return TraceContext()
        if not all(char in "0123456789abcdef" for char in trace_id.lower()):
            return TraceContext()
        if not all(char in "0123456789abcdef" for char in parent_span_id.lower()):
            return TraceContext()
        return TraceContext(trace_id=trace_id.lower(), parent_span_id=parent_span_id.lower())

    def to_headers(self) -> dict[str, str]:
        span_id = uuid.uuid4().hex[:16]
        parent = self.parent_span_id or span_id
        return {"traceparent": f"00-{self.trace_id}-{parent}-01"}


class OTelEmitter(EventEmitter):
    """Converts DecisionEvents into OTel-compatible span data."""

    def __init__(self, path: str = ".orchesis/traces.jsonl"):
        self.path = Path(path)

    def emit(self, event: DecisionEvent) -> None:
        span = self._decision_to_span(event)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open("a", encoding="utf-8") as file:
            file.write(json.dumps(asdict(span), ensure_ascii=False) + "\n")

    def _decision_to_span(self, event: DecisionEvent) -> SpanData:
        end_time_ns = time.time_ns()
        duration_us = max(0, int(event.evaluation_duration_us))
        start_time_ns = end_time_ns - (duration_us * 1000)
        snapshot = event.state_snapshot if isinstance(event.state_snapshot, dict) else {}
        trace_id = (
            snapshot.get("trace_id")
            if isinstance(snapshot.get("trace_id"), str)
            else uuid.uuid4().hex
        )
        parent_span_id = (
            snapshot.get("parent_span_id")
            if isinstance(snapshot.get("parent_span_id"), str)
            else None
        )
        span_id = uuid.uuid4().hex[:16]
        rules_triggered = set(event.rules_triggered)
        events = [
            {
                "name": "rule.checked",
                "timestamp_ns": end_time_ns,
                "attributes": {
                    "rule_name": rule_name,
                    "triggered": rule_name in rules_triggered,
                },
            }
            for rule_name in event.rules_checked
        ]
        attributes: dict[str, str | int | float | bool] = {
            "orchesis.agent_id": event.agent_id,
            "orchesis.tool": event.tool,
            "orchesis.decision": event.decision,
            "orchesis.policy_version": event.policy_version,
            "orchesis.evaluation_duration_us": event.evaluation_duration_us,
            "orchesis.rules_checked_count": len(event.rules_checked),
            "orchesis.reasons_count": len(event.reasons),
        }
        return SpanData(
            trace_id=trace_id,
            span_id=span_id,
            parent_span_id=parent_span_id,
            operation="orchesis.evaluate",
            start_time_ns=start_time_ns,
            end_time_ns=end_time_ns,
            status="OK" if event.decision == "ALLOW" else "ERROR",
            attributes=attributes,
            events=events,
        )


class ProxySpanEmitter:
    """
    Creates spans for each proxy pipeline phase.
    Feeds to OTLPSpanExporter + JSONL file.

    Usage from proxy:
        span = emitter.start_span("orchesis.phase.cascade", trace_ctx)
        # ... do cascade work ...
        emitter.end_span(span, attributes={...})
    """

    def __init__(
        self,
        jsonl_path: str = ".orchesis/traces.jsonl",
        otlp_exporter: OTLPSpanExporter | None = None,
    ) -> None:
        self._jsonl_emitter = OTelEmitter(path=jsonl_path)
        self._otlp_exporter = otlp_exporter

    def start_span(
        self,
        operation: str,
        trace_ctx: TraceContext,
        parent_span_id: str | None = None,
        attributes: dict[str, str | int | float | bool] | None = None,
    ) -> SpanData:
        """Create a new span (not yet ended)."""
        return SpanData(
            trace_id=trace_ctx.trace_id,
            span_id=uuid.uuid4().hex[:16],
            parent_span_id=parent_span_id or trace_ctx.parent_span_id,
            operation=operation,
            start_time_ns=time.time_ns(),
            end_time_ns=0,
            status="OK",
            attributes=dict(attributes or {}),
            events=[],
        )

    def end_span(
        self,
        span: SpanData,
        status: str = "OK",
        attributes: dict[str, str | int | float | bool] | None = None,
    ) -> None:
        """End span and export to all destinations."""
        finished = SpanData(
            trace_id=span.trace_id,
            span_id=span.span_id,
            parent_span_id=span.parent_span_id,
            operation=span.operation,
            start_time_ns=span.start_time_ns,
            end_time_ns=time.time_ns(),
            status=status,
            attributes={**span.attributes, **(attributes or {})},
            events=span.events,
        )
        if self._otlp_exporter:
            self._otlp_exporter.export_span(finished)

    def create_request_span(
        self,
        trace_ctx: TraceContext,
        model: str,
        provider: str,
        session_id: str = "",
        agent_id: str = "",
    ) -> SpanData:
        """Create the root span for an LLM request with GenAI attrs."""
        return self.start_span(
            operation="orchesis.proxy.request",
            trace_ctx=trace_ctx,
            attributes={
                "gen_ai.system": provider,
                "gen_ai.operation.name": "chat",
                "gen_ai.request.model": model,
                "orchesis.session_id": session_id,
                "orchesis.agent_id": agent_id,
            },
        )

    def create_phase_span(
        self,
        phase_name: str,
        trace_ctx: TraceContext,
        parent_span_id: str,
    ) -> SpanData:
        """Create a child span for a proxy phase."""
        return self.start_span(
            operation=f"orchesis.phase.{phase_name}",
            trace_ctx=trace_ctx,
            parent_span_id=parent_span_id,
            attributes={"orchesis.proxy.phase": phase_name},
        )



"""Incident forensics engine over decision telemetry logs."""

from __future__ import annotations

import json
import hashlib
from collections import Counter, defaultdict
from dataclasses import asdict, dataclass, field
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any

from orchesis.replay import read_events_from_jsonl
from orchesis.telemetry import DecisionEvent


def _now_iso() -> str:
    """Return current UTC timestamp in ISO format."""
    return datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")


def _parse_ts(value: str) -> datetime | None:
    """Parse timestamp into timezone-aware UTC datetime."""
    try:
        parsed = datetime.fromisoformat(value.replace("Z", "+00:00"))
    except ValueError:
        return None
    if parsed.tzinfo is None:
        return parsed.replace(tzinfo=timezone.utc)
    return parsed.astimezone(timezone.utc)


@dataclass
class Incident:
    """Forensics incident model."""

    id: str
    severity: str
    type: str
    title: str
    timestamp: str
    agent_id: str | None = None
    tool: str | None = None
    details: dict[str, Any] = field(default_factory=dict)
    related_events: list[dict[str, Any]] = field(default_factory=list)
    policy_version: str | None = None
    resolution: str | None = None
    timeline: list[dict[str, Any]] = field(default_factory=list)


@dataclass
class IncidentReport:
    """Aggregate report for incidents in a period."""

    incidents: list[Incident]
    summary: dict[str, Any]
    period_start: str
    period_end: str
    mean_time_to_detect: float | None = None


class ForensicsEngine:
    """Analyze audit logs and produce incident reports."""

    def __init__(
        self,
        decisions_path: str = ".orchesis/decisions.jsonl",
        fuzz_meta_path: str = ".orchesis/fuzz_meta.json",
    ) -> None:
        self._decisions_path = Path(decisions_path)
        self._fuzz_meta_path = Path(fuzz_meta_path)
        self._incident_index: dict[str, Incident] = {}

    def _load_events(self) -> list[DecisionEvent]:
        return read_events_from_jsonl(self._decisions_path)

    def _new_incident(
        self,
        *,
        severity: str,
        incident_type: str,
        title: str,
        event: DecisionEvent,
        details: dict[str, Any] | None = None,
        related_events: list[DecisionEvent] | None = None,
    ) -> Incident:
        inc = Incident(
            id=(
                "inc-"
                + hashlib.sha256(
                    f"{event.event_id}:{incident_type}:{title}:{event.timestamp}".encode("utf-8")
                ).hexdigest()[:8]
            ),
            severity=severity,
            type=incident_type,
            title=title,
            timestamp=event.timestamp,
            agent_id=event.agent_id,
            tool=event.tool,
            details=details or {},
            related_events=[asdict(item) for item in (related_events or [event])],
            policy_version=event.policy_version,
            resolution="open",
            timeline=[{"ts": event.timestamp, "action": "detected"}],
        )
        return inc

    def detect_incidents(
        self,
        since: str | None = None,
        severity_filter: str | None = None,
    ) -> list[Incident]:
        """Scan decision log for incidents."""
        events = self._load_events()
        threshold = _parse_ts(since) if isinstance(since, str) else None
        filtered: list[DecisionEvent] = []
        for event in events:
            ts = _parse_ts(event.timestamp)
            if threshold is not None and (ts is None or ts < threshold):
                continue
            filtered.append(event)

        if not filtered:
            self._incident_index = {}
            return []

        incidents: list[Incident] = []
        avg_latency = sum(item.evaluation_duration_us for item in filtered) / max(1, len(filtered))
        deny_windows: dict[str, list[datetime]] = defaultdict(list)
        seen_burst_keys: set[tuple[str, str]] = set()

        for event in filtered:
            reasons_l = [reason.lower() for reason in event.reasons]
            ts = _parse_ts(event.timestamp) or datetime.now(timezone.utc)

            # 1) bypass marker
            if event.decision == "DENY" and any("bypass" in reason for reason in reasons_l):
                incidents.append(
                    self._new_incident(
                        severity="critical",
                        incident_type="bypass",
                        title="Bypass attempt detected",
                        event=event,
                        details={"reasons": event.reasons},
                    )
                )

            # 2) brute force denies in one minute
            if event.decision == "DENY":
                agent_window = deny_windows[event.agent_id]
                agent_window.append(ts)
                one_minute_ago = ts - timedelta(minutes=1)
                deny_windows[event.agent_id] = [item for item in agent_window if item >= one_minute_ago]
                if len(deny_windows[event.agent_id]) > 5:
                    key = (event.agent_id, ts.strftime("%Y-%m-%dT%H:%M"))
                    if key not in seen_burst_keys:
                        seen_burst_keys.add(key)
                        incidents.append(
                            self._new_incident(
                                severity="high",
                                incident_type="anomaly",
                                title="Brute force deny burst",
                                event=event,
                                details={"deny_count_1m": len(deny_windows[event.agent_id])},
                            )
                        )

            # 3) privilege escalation attempts
            if any("lacks capability" in reason or "not in allowed_tools" in reason for reason in reasons_l):
                incidents.append(
                    self._new_incident(
                        severity="medium",
                        incident_type="identity_violation",
                        title="Privilege escalation attempt",
                        event=event,
                    )
                )

            # 4) budget exceeded
            if any("daily budget exceeded" in reason for reason in reasons_l):
                incidents.append(
                    self._new_incident(
                        severity="high",
                        incident_type="budget_breach",
                        title="Budget limit exceeded",
                        event=event,
                    )
                )

            # 5) rate limit exceeded
            if any("rate_limit" in reason and "exceeded" in reason for reason in reasons_l):
                incidents.append(
                    self._new_incident(
                        severity="medium",
                        incident_type="rate_limit_breach",
                        title="Rate limit exceeded",
                        event=event,
                    )
                )

            # 6) unknown agent
            if event.agent_id in {"__global__", "__unknown__", "unknown"} or event.agent_id.startswith(
                "unknown"
            ):
                incidents.append(
                    self._new_incident(
                        severity="medium",
                        incident_type="identity_violation",
                        title="Unknown agent identity observed",
                        event=event,
                    )
                )

            # 7) latency anomaly
            if avg_latency > 0 and event.evaluation_duration_us > avg_latency * 10:
                incidents.append(
                    self._new_incident(
                        severity="low",
                        incident_type="anomaly",
                        title="Latency spike detected",
                        event=event,
                        details={
                            "latency_us": event.evaluation_duration_us,
                            "average_us": round(avg_latency, 2),
                        },
                    )
                )

        if severity_filter is not None:
            target = severity_filter.strip().lower()
            incidents = [item for item in incidents if item.severity == target]

        self._incident_index = {item.id: item for item in incidents}
        incidents.sort(key=lambda item: item.timestamp, reverse=True)
        return incidents

    def get_incident(self, incident_id: str) -> Incident | None:
        """Get single incident with context."""
        if incident_id in self._incident_index:
            return self._incident_index[incident_id]
        for item in self.detect_incidents():
            if item.id == incident_id:
                return item
        return None

    def build_report(
        self,
        since: str | None = None,
        until: str | None = None,
    ) -> IncidentReport:
        """Build comprehensive report for incident period."""
        incidents = self.detect_incidents(since=since)
        end_dt = _parse_ts(until) if isinstance(until, str) else datetime.now(timezone.utc)
        if end_dt is None:
            end_dt = datetime.now(timezone.utc)
        if until is not None:
            incidents = [item for item in incidents if (_parse_ts(item.timestamp) or end_dt) <= end_dt]

        by_severity = Counter(item.severity for item in incidents)
        by_type = Counter(item.type for item in incidents)
        by_agent = Counter(item.agent_id or "unknown" for item in incidents)

        if incidents:
            detect_delays: list[float] = []
            for item in incidents:
                if item.timeline:
                    first = _parse_ts(item.timeline[0]["ts"]) if isinstance(item.timeline[0], dict) else None
                    detected = _parse_ts(item.timestamp)
                    if first is not None and detected is not None:
                        detect_delays.append(max(0.0, (detected - first).total_seconds()))
            mttd = (sum(detect_delays) / len(detect_delays)) if detect_delays else 0.0
            period_start = min(item.timestamp for item in incidents)
            period_end = max(item.timestamp for item in incidents)
        else:
            mttd = None
            period_start = since or _now_iso()
            period_end = until or _now_iso()

        summary = {
            "total": len(incidents),
            "by_severity": dict(by_severity),
            "by_type": dict(by_type),
            "by_agent": dict(by_agent),
        }
        return IncidentReport(
            incidents=incidents,
            summary=summary,
            period_start=period_start,
            period_end=period_end,
            mean_time_to_detect=mttd,
        )

    def agent_risk_profile(self, agent_id: str) -> dict[str, Any]:
        """Compute risk profile for one agent."""
        events = [item for item in self._load_events() if item.agent_id == agent_id]
        total = len(events)
        denied = sum(1 for item in events if item.decision == "DENY")
        deny_rate = (denied / total) if total else 0.0
        incidents = [item for item in self.detect_incidents() if item.agent_id == agent_id]
        incident_density = (len(incidents) / total) if total else 0.0
        risk_score = min(1.0, (deny_rate * 0.6) + (incident_density * 0.4))
        denied_events = [item for item in events if item.decision == "DENY"]
        top_tools = Counter(item.tool for item in denied_events).most_common(5)
        reasons = Counter(reason for item in denied_events for reason in item.reasons).most_common(5)

        now = datetime.now(timezone.utc)
        last_24 = [item for item in events if (_parse_ts(item.timestamp) or now) >= now - timedelta(hours=24)]
        prev_24 = [
            item
            for item in events
            if now - timedelta(hours=48) <= (_parse_ts(item.timestamp) or now) < now - timedelta(hours=24)
        ]
        last_deny_rate = (
            sum(1 for item in last_24 if item.decision == "DENY") / len(last_24) if last_24 else 0.0
        )
        prev_deny_rate = (
            sum(1 for item in prev_24 if item.decision == "DENY") / len(prev_24) if prev_24 else 0.0
        )
        trend = "stable"
        if last_deny_rate > prev_deny_rate + 0.02:
            trend = "increasing"
        elif last_deny_rate + 0.02 < prev_deny_rate:
            trend = "decreasing"

        return {
            "agent_id": agent_id,
            "total_requests": total,
            "denied": denied,
            "deny_rate": round(deny_rate, 4),
            "top_denied_tools": top_tools,
            "top_deny_reasons": reasons,
            "incidents": len(incidents),
            "risk_score": round(risk_score, 4),
            "trend": trend,
        }

    def attack_timeline(
        self,
        incident_id: str | None = None,
        agent_id: str | None = None,
        last_n: int = 50,
    ) -> list[dict[str, Any]]:
        """Build chronological event timeline for visualization."""
        events = self._load_events()
        if incident_id is not None:
            incident = self.get_incident(incident_id)
            if incident is None:
                return []
            agent_id = incident.agent_id
        if agent_id is not None:
            events = [item for item in events if item.agent_id == agent_id]
        events = sorted(events, key=lambda item: item.timestamp)
        if last_n > 0:
            events = events[-last_n:]
        return [
            {
                "ts": item.timestamp,
                "agent_id": item.agent_id,
                "tool": item.tool,
                "decision": item.decision,
                "reasons": list(item.reasons),
                "latency_us": item.evaluation_duration_us,
            }
            for item in events
        ]

    def export_json(self, report: IncidentReport) -> str:
        """Export report as JSON string."""
        payload = {
            "incidents": [asdict(item) for item in report.incidents],
            "summary": report.summary,
            "period_start": report.period_start,
            "period_end": report.period_end,
            "mean_time_to_detect": report.mean_time_to_detect,
        }
        return json.dumps(payload, ensure_ascii=False, indent=2)

    def export_markdown(self, report: IncidentReport) -> str:
        """Export report as markdown."""
        sev = report.summary.get("by_severity", {})
        most_affected = "n/a"
        by_agent = report.summary.get("by_agent", {})
        if isinstance(by_agent, dict) and by_agent:
            most_affected = max(by_agent.items(), key=lambda item: item[1])[0]
        lines = [
            f"# Incident Report: {report.period_start} to {report.period_end}",
            "",
            "## Summary",
            f"- Total incidents: {report.summary.get('total', 0)}",
            (
                "- Critical: "
                f"{sev.get('critical', 0)}, High: {sev.get('high', 0)}, "
                f"Medium: {sev.get('medium', 0)}, Low: {sev.get('low', 0)}"
            ),
            f"- Most affected agent: {most_affected}",
            "",
            "## Incidents",
            "",
        ]
        for item in report.incidents:
            lines.extend(
                [
                    f"### [{item.id.upper()}] {item.severity.title()}: {item.title}",
                    f"**Time:** {item.timestamp}",
                    f"**Agent:** {item.agent_id or 'unknown'}",
                    f"**Tool:** {item.tool or 'unknown'}",
                    f"**Details:** {item.details or {}}",
                    f"**Status:** {item.resolution or 'open'}",
                    "",
                ]
            )
        lines.extend(
            [
                "## Agent Risk Profiles",
                "",
                "| Agent | Requests | Denied | Deny Rate | Risk Score |",
                "|-------|----------|--------|-----------|------------|",
            ]
        )
        if isinstance(by_agent, dict):
            for agent in sorted(by_agent):
                profile = self.agent_risk_profile(agent)
                lines.append(
                    f"| {agent} | {profile['total_requests']} | {profile['denied']} | "
                    f"{profile['deny_rate']*100:.1f}% | {profile['risk_score']:.2f} |"
                )
        return "\n".join(lines)



from __future__ import annotations

from pathlib import Path

import yaml


ROOT = Path(__file__).resolve().parents[1]


def test_contributing_exists() -> None:
    assert (ROOT / "CONTRIBUTING.md").exists()


def test_contributing_has_sections() -> None:
    content = (ROOT / "CONTRIBUTING.md").read_text(encoding="utf-8")
    assert "## Getting Started" in content
    assert "## Development Workflow" in content
    assert "## Testing" in content
    assert "## Adding a Plugin" in content


def test_security_has_disclosure() -> None:
    content = (ROOT / "docs" / "SECURITY.md").read_text(encoding="utf-8")
    assert "## Reporting Vulnerabilities" in content


def test_issue_templates_exist() -> None:
    issue_dir = ROOT / ".github" / "ISSUE_TEMPLATE"
    assert (issue_dir / "bug_report.md").exists()
    assert (issue_dir / "feature_request.md").exists()
    assert (issue_dir / "bypass_report.md").exists()


def test_pr_template_exists() -> None:
    assert (ROOT / ".github" / "PULL_REQUEST_TEMPLATE.md").exists()


def test_ci_workflow_valid_yaml() -> None:
    path = ROOT / ".github" / "workflows" / "ci.yml"
    data = yaml.safe_load(path.read_text(encoding="utf-8"))
    assert isinstance(data, dict)


def test_ci_tests_python_312_only() -> None:
    content = (ROOT / ".github" / "workflows" / "ci.yml").read_text(encoding="utf-8")
    assert '"3.12"' in content
    assert '"3.11"' not in content
    assert '"3.13"' not in content


def test_readme_has_badges() -> None:
    content = (ROOT / "README.md").read_text(encoding="utf-8")
    assert '<img src="docs/banner.svg" alt="Orchesis" />' in content
    assert "img.shields.io/pypi/v/orchesis" in content
    assert "img.shields.io/badge/tests-2738%20passing-brightgreen" in content


def test_readme_has_docker_section() -> None:
    content = (ROOT / "README.md").read_text(encoding="utf-8")
    assert "## Installation" in content
    assert "pip install orchesis" in content
    assert "orchesis quickstart --preset openclaw" in content


def test_readme_has_project_status() -> None:
    content = (ROOT / "README.md").read_text(encoding="utf-8")
    assert "## By the numbers" in content
    assert "2,738" in content
    assert "## What Orchesis does" in content



from __future__ import annotations

import time
import tracemalloc
from concurrent.futures import ThreadPoolExecutor
from random import Random

from orchesis.engine import evaluate
from orchesis.identity import AgentIdentity, AgentRegistry, TrustTier
from orchesis.state import RateLimitTracker
from orchesis.telemetry import InMemoryEmitter


def _policy(*, rate_limit: int = 10, daily_budget: float = 1000.0, max_cost: float = 10.0) -> dict:
    return {
        "rules": [
            {"name": "budget_limit", "max_cost_per_call": max_cost, "daily_budget": daily_budget},
            {"name": "rate_limit", "max_requests_per_minute": rate_limit},
            {"name": "file_access", "allowed_paths": ["/data"], "denied_paths": ["/etc", "/root"]},
            {"name": "sql_restriction", "denied_operations": ["DROP", "DELETE"]},
        ]
    }


def test_1000_agents_concurrent() -> None:
    policy = _policy(rate_limit=10)
    tracker = RateLimitTracker(persist_path=None)

    def _call(i: int) -> bool:
        agent_id = f"agent-{i // 10}"
        request = {
            "tool": "read_file",
            "params": {"path": "/data/safe.txt"},
            "cost": 0.1,
            "context": {"agent": agent_id},
        }
        decision = evaluate(request, policy, state=tracker)
        return decision.allowed

    with ThreadPoolExecutor(max_workers=100) as pool:
        results = list(pool.map(_call, range(10_000)))
    assert all(isinstance(item, bool) for item in results)

    # Every agent sends exactly 10 requests against rate_limit=10 => all should pass.
    for agent_idx in range(1000):
        allowed = sum(results[agent_idx * 10 : (agent_idx + 1) * 10])
        assert allowed == 10


def test_5000_concurrent_evaluations() -> None:
    policy = _policy(rate_limit=500)
    tracker = RateLimitTracker(persist_path=None)
    rng = Random(42)
    requests: list[tuple[dict, bool]] = []
    for i in range(5000):
        roll = rng.random()
        base = {"context": {"agent": f"agent-{i % 50}"}, "cost": 0.1}
        if roll < 0.60:
            req = {"tool": "read_file", "params": {"path": "/data/report.txt"}, **base}
            requests.append((req, True))
        elif roll < 0.80:
            req = {"tool": "read_file", "params": {"path": "/etc/passwd"}, **base}
            requests.append((req, False))
        else:
            req = {"tool": "run_sql", "params": {"query": "DROP TABLE users"}, **base}
            requests.append((req, False))

    started = time.perf_counter()

    def _call(item: tuple[dict, bool]) -> tuple[bool, bool]:
        req, expected_allow = item
        decision = evaluate(req, policy, state=tracker)
        return decision.allowed, expected_allow

    with ThreadPoolExecutor(max_workers=200) as pool:
        results = list(pool.map(_call, requests))
    elapsed = time.perf_counter() - started

    for allowed, expected_allow in results:
        assert allowed == expected_allow
    assert elapsed < 60.0, f"5000 concurrent evaluations took {elapsed:.1f}s (limit 60s)"


def test_rate_limit_exact_under_concurrency() -> None:
    policy = _policy(rate_limit=10)
    tracker = RateLimitTracker(persist_path=None)

    def _call(i: int) -> tuple[str, bool]:
        agent_id = f"agent-{i % 20}"
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/safe.txt"},
                "cost": 0.1,
                "context": {"agent": agent_id},
            },
            policy,
            state=tracker,
        )
        return agent_id, decision.allowed

    with ThreadPoolExecutor(max_workers=200) as pool:
        results = list(pool.map(_call, range(400)))

    by_agent: dict[str, list[bool]] = {}
    for agent_id, allowed in results:
        by_agent.setdefault(agent_id, []).append(allowed)

    for values in by_agent.values():
        assert sum(values) == 10
        assert len(values) - sum(values) == 10


def test_budget_exact_under_concurrency() -> None:
    policy = _policy(rate_limit=1000, daily_budget=5.0, max_cost=10.0)
    tracker = RateLimitTracker(persist_path=None)

    def _call(i: int) -> tuple[str, bool]:
        agent_id = f"agent-{i % 10}"
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/safe.txt"},
                "cost": 1.0,
                "context": {"agent": agent_id},
            },
            policy,
            state=tracker,
        )
        return agent_id, decision.allowed

    with ThreadPoolExecutor(max_workers=100) as pool:
        results = list(pool.map(_call, range(100)))

    by_agent: dict[str, list[bool]] = {}
    for agent_id, allowed in results:
        by_agent.setdefault(agent_id, []).append(allowed)

    for agent_id, values in by_agent.items():
        assert sum(values) == 5
        assert len(values) - sum(values) == 5
        assert tracker.get_agent_budget_spent(agent_id, window_seconds=86400) == 5.0


def test_session_isolation_under_load() -> None:
    policy = _policy(rate_limit=10)
    tracker = RateLimitTracker(persist_path=None)

    def _call(i: int) -> tuple[str, str, bool]:
        agent = f"agent-{(i // 200) % 5}"
        session = f"session-{(i // 20) % 10}"
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/safe.txt"},
                "cost": 0.1,
                "context": {"agent": agent, "session": session},
            },
            policy,
            state=tracker,
        )
        return agent, session, decision.allowed

    with ThreadPoolExecutor(max_workers=100) as pool:
        results = list(pool.map(_call, range(1000)))

    by_key: dict[tuple[str, str], list[bool]] = {}
    for agent, session, allowed in results:
        by_key.setdefault((agent, session), []).append(allowed)

    for values in by_key.values():
        assert sum(values) == 10
        assert len(values) - sum(values) == 10


def test_mixed_identity_tiers_concurrent() -> None:
    policy = _policy(rate_limit=1000)
    registry = AgentRegistry(
        agents={
            "blocked": AgentIdentity("blocked", "blocked", trust_tier=TrustTier.BLOCKED),
            "intern": AgentIdentity("intern", "intern", trust_tier=TrustTier.INTERN),
            "operator": AgentIdentity("operator", "operator", trust_tier=TrustTier.OPERATOR),
            "principal": AgentIdentity("principal", "principal", trust_tier=TrustTier.PRINCIPAL),
        }
    )
    tracker = RateLimitTracker(persist_path=None)

    calls: list[tuple[dict, bool]] = []
    for _ in range(100):
        calls.append(
            (
                {
                    "tool": "read_file",
                    "params": {"path": "/data/a.txt"},
                    "cost": 0.1,
                    "context": {"agent": "blocked"},
                },
                False,
            )
        )
        calls.append(
            (
                {
                    "tool": "read_file",
                    "params": {"path": "/data/a.txt"},
                    "cost": 0.1,
                    "context": {"agent": "intern"},
                },
                True,
            )
        )
        calls.append(
            (
                {
                    "tool": "write_file",
                    "params": {"path": "/data/a.txt"},
                    "cost": 0.1,
                    "context": {"agent": "intern"},
                },
                False,
            )
        )
        calls.append(
            (
                {
                    "tool": "write_file",
                    "params": {"path": "/data/a.txt"},
                    "cost": 0.1,
                    "context": {"agent": "operator"},
                },
                True,
            )
        )
        calls.append(
            (
                {
                    "tool": "delete_file",
                    "params": {"path": "/data/a.txt"},
                    "cost": 0.1,
                    "context": {"agent": "principal"},
                },
                True,
            )
        )

    def _call(item: tuple[dict, bool]) -> bool:
        req, expected = item
        decision = evaluate(req, policy, state=tracker, registry=registry)
        return decision.allowed == expected

    with ThreadPoolExecutor(max_workers=100) as pool:
        results = list(pool.map(_call, calls))
    assert all(results)


def test_policy_reload_during_traffic() -> None:
    policy_a = _policy(rate_limit=1000, max_cost=1.0, daily_budget=100.0)
    policy_b = _policy(rate_limit=1000, max_cost=0.5, daily_budget=100.0)
    current = [policy_a]
    tracker = RateLimitTracker(persist_path=None)

    def _call(i: int) -> bool:
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/safe.txt"},
                "cost": 0.75,
                "context": {"agent": f"agent-{i % 20}"},
            },
            current[0],
            state=tracker,
        )
        return decision.allowed

    with ThreadPoolExecutor(max_workers=100) as pool:
        first_half = list(pool.map(_call, range(500)))
        current[0] = policy_b
        second_half = list(pool.map(_call, range(500, 1000)))

    assert all(first_half)
    assert not any(second_half)


def test_state_persistence_under_crash(tmp_path) -> None:  # noqa: ANN001
    state_file = tmp_path / "state.jsonl"
    policy = _policy(rate_limit=2000, daily_budget=2000.0)

    tracker1 = RateLimitTracker(persist_path=state_file)
    for i in range(500):
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/a.txt"},
                "cost": 1.0,
                "context": {"agent": "persist-agent"},
            },
            policy,
            state=tracker1,
        )
        assert decision.allowed
    tracker1.flush()

    tracker2 = RateLimitTracker(persist_path=state_file)
    assert tracker2.get_count("read_file", 86400, agent_id="persist-agent") >= 500
    assert tracker2.get_agent_budget_spent("persist-agent", 86400) >= 500.0

    for _ in range(500):
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/a.txt"},
                "cost": 1.0,
                "context": {"agent": "persist-agent"},
            },
            policy,
            state=tracker2,
        )
        assert decision.allowed
    assert tracker2.get_count("read_file", 86400, agent_id="persist-agent") >= 1000
    assert tracker2.get_agent_budget_spent("persist-agent", 86400) >= 1000.0


def test_sustained_throughput_60_seconds() -> None:
    policy = _policy(rate_limit=10_000_000, daily_budget=100000.0)
    tracker = RateLimitTracker(persist_path=None)
    latencies_us: list[int] = []

    tracemalloc.start()
    start_mem, _ = tracemalloc.get_traced_memory()
    started = time.perf_counter()
    total = 0
    while time.perf_counter() - started < 60:
        before = time.perf_counter_ns()
        decision = evaluate(
            {
                "tool": "read_file",
                "params": {"path": "/data/load.txt"},
                "cost": 0.1,
                "context": {"agent": "sustain"},
            },
            policy,
            state=tracker,
        )
        latencies_us.append(max(0, (time.perf_counter_ns() - before) // 1000))
        total += 1
        assert decision.allowed

    end_mem, _ = tracemalloc.get_traced_memory()
    tracemalloc.stop()

    elapsed = max(0.001, time.perf_counter() - started)
    throughput = total / elapsed
    sorted_lat = sorted(latencies_us)
    p99_idx = max(0, int(0.99 * (len(sorted_lat) - 1)))
    p99 = sorted_lat[p99_idx]
    memory_growth_mb = max(0.0, (end_mem - start_mem) / (1024.0 * 1024.0))
    assert throughput >= 100.0
    assert p99 < 20_000
    assert memory_growth_mb <= 100.0


def test_telemetry_doesnt_slow_evaluation() -> None:
    policy = {
        "rules": [
            {"name": "budget_limit", "max_cost_per_call": 10.0},
            {
                "name": "regex_guard",
                "type": "regex_match",
                "field": "params.query",
                "deny_patterns": [f"NEVER_MATCH_{idx}$" for idx in range(1200)],
            },
        ]
    }
    request = {
        "tool": "read_file",
        "params": {
            "path": "/data/telemetry.txt",
            "query": "SELECT * FROM telemetry_table WHERE x=1",
        },
        "cost": 0.0,
        "context": {"agent": "telemetry-agent"},
    }

    tracker_without = RateLimitTracker(persist_path=None)
    started_without = time.perf_counter()
    for _ in range(200):
        evaluate(request, policy, state=tracker_without)
    elapsed_without = max(1e-9, time.perf_counter() - started_without)

    tracker_with = RateLimitTracker(persist_path=None)
    emitter = InMemoryEmitter()
    started_with = time.perf_counter()
    for _ in range(200):
        evaluate(request, policy, state=tracker_with, emitter=emitter)
    elapsed_with = time.perf_counter() - started_with

    overhead = (elapsed_with - elapsed_without) / elapsed_without
    # Perf noise on shared CI/VMs can be spiky; keep a strict but stable threshold.
    assert overhead < 0.60



from __future__ import annotations

import threading
import time

from orchesis.session_risk import RiskSignal, SessionRiskAccumulator


def _sig(
    category: str = "prompt_injection",
    *,
    confidence: float = 1.0,
    severity: str = "low",
    source: str = "threat_intel",
    description: str = "signal",
    timestamp: float = 0.0,
) -> RiskSignal:
    return RiskSignal(
        category=category,
        confidence=confidence,
        severity=severity,
        source=source,
        description=description,
        timestamp=timestamp,
    )


def test_empty_session_allows() -> None:
    acc = SessionRiskAccumulator()
    out = acc.evaluate("s1")
    assert out.action == "allow"
    assert out.composite_score == 0.0


def test_single_low_signal_allows() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig(severity="low", confidence=0.5))
    out = acc.evaluate("s1")
    assert out.action == "allow"
    assert out.composite_score < 30.0


def test_single_critical_signal_warns_or_blocks() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig(severity="critical", confidence=1.0))
    out = acc.evaluate("s1")
    assert out.action in {"warn", "block"}


def test_multiple_low_signals_same_category_accumulates() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig(category="prompt_injection", severity="low", confidence=1.0))
    acc.record_signal("s1", _sig(category="prompt_injection", severity="low", confidence=1.0))
    out = acc.evaluate("s1")
    assert out.composite_score >= 10.0


def test_multiple_signals_different_categories_diversity_bonus() -> None:
    acc = SessionRiskAccumulator(category_diversity_bonus=10.0)
    acc.record_signal("s1", _sig(category="prompt_injection", severity="low"))
    acc.record_signal("s1", _sig(category="data_exfiltration", severity="low"))
    out = acc.evaluate("s1")
    assert out.unique_categories == 2
    assert out.composite_score >= 20.0


def test_three_low_different_categories_crosses_warn() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig(category="prompt_injection", severity="low"))
    acc.record_signal("s1", _sig(category="data_exfiltration", severity="low"))
    acc.record_signal("s1", _sig(category="path_traversal", severity="low"))
    out = acc.evaluate("s1")
    assert out.action in {"warn", "block"}
    assert out.composite_score >= 30.0


def test_escalation_observe_to_warn() -> None:
    acc = SessionRiskAccumulator(warn_threshold=10.0, block_threshold=80.0)
    acc.record_signal("s1", _sig(category="a", severity="medium", confidence=1.0))
    out = acc.evaluate("s1")
    assert out.action == "warn"


def test_escalation_warn_to_block() -> None:
    acc = SessionRiskAccumulator(warn_threshold=20.0, block_threshold=40.0)
    acc.record_signal("s1", _sig(category="a", severity="high", confidence=1.0))
    acc.record_signal("s1", _sig(category="b", severity="high", confidence=1.0))
    out = acc.evaluate("s1")
    assert out.action == "block"


def test_escalation_does_not_downgrade_without_decay() -> None:
    acc = SessionRiskAccumulator(warn_threshold=20.0, block_threshold=40.0)
    acc.record_signal("s1", _sig(category="a", severity="high"))
    acc.record_signal("s1", _sig(category="b", severity="high"))
    first = acc.evaluate("s1")
    second = acc.evaluate("s1")
    assert first.action == "block"
    assert second.action == "block"


def test_old_signals_decay_reduces_score() -> None:
    now = time.monotonic()
    acc = SessionRiskAccumulator(decay_half_life_seconds=10.0)
    acc.record_signal("s1", _sig(category="a", severity="high", timestamp=now - 60.0))
    out = acc.evaluate("s1")
    assert out.composite_score < 5.0


def test_fresh_signals_full_weight() -> None:
    acc = SessionRiskAccumulator(decay_half_life_seconds=300.0)
    acc.record_signal("s1", _sig(category="a", severity="medium", confidence=1.0))
    out = acc.evaluate("s1")
    assert 10.0 <= out.composite_score <= 13.0


def test_half_life_halves_weight() -> None:
    now = time.monotonic()
    acc = SessionRiskAccumulator(decay_half_life_seconds=10.0)
    acc.record_signal("s1", _sig(category="a", severity="medium", confidence=1.0, timestamp=now - 10.0))
    out = acc.evaluate("s1")
    assert 5.0 <= out.composite_score <= 7.0


def test_session_ttl_cleanup() -> None:
    acc = SessionRiskAccumulator(session_ttl_seconds=60.0)
    acc.record_signal("s1", _sig(category="a"))
    with acc._lock:  # noqa: SLF001
        acc._sessions["s1"].last_updated = time.monotonic() - 1000.0  # noqa: SLF001
    _ = acc.evaluate("another-session")
    assert acc.get_session_state("s1") is None


def test_max_signals_per_session_bounded() -> None:
    acc = SessionRiskAccumulator(max_signals_per_session=10)
    for i in range(30):
        acc.record_signal("s1", _sig(category=f"c{i}", severity="low"))
    state = acc.get_session_state("s1")
    assert state is not None
    assert state["total_signals"] == 10


def test_reset_session() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig())
    assert acc.reset_session("s1") is True
    assert acc.get_session_state("s1") is None


def test_get_session_state() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig(category="x"))
    state = acc.get_session_state("s1")
    assert state is not None
    assert state["session_id"] == "s1"
    assert state["total_signals"] == 1


def test_concurrent_record_and_evaluate() -> None:
    acc = SessionRiskAccumulator()
    stop = False
    errors: list[Exception] = []

    def writer() -> None:
        nonlocal stop
        try:
            while not stop:
                acc.record_signal("s1", _sig(category="c", severity="low", confidence=0.7))
        except Exception as exc:  # noqa: BLE001
            errors.append(exc)

    def reader() -> None:
        nonlocal stop
        try:
            for _ in range(500):
                _ = acc.evaluate("s1")
            stop = True
        except Exception as exc:  # noqa: BLE001
            errors.append(exc)

    t1 = threading.Thread(target=writer)
    t2 = threading.Thread(target=reader)
    t1.start()
    t2.start()
    t1.join(timeout=2.0)
    t2.join(timeout=2.0)
    assert not errors
    assert acc.stats["signals_recorded"] > 0


def test_disabled_always_allows() -> None:
    acc = SessionRiskAccumulator(enabled=False)
    acc.record_signal("s1", _sig(severity="critical"))
    out = acc.evaluate("s1")
    assert out.action == "allow"
    assert out.reason == "disabled"


def test_custom_thresholds() -> None:
    acc = SessionRiskAccumulator(warn_threshold=10.0, block_threshold=20.0)
    acc.record_signal("s1", _sig(category="a", severity="medium"))
    out = acc.evaluate("s1")
    assert out.action in {"warn", "block"}


def test_default_config() -> None:
    acc = SessionRiskAccumulator()
    stats = acc.stats
    assert stats["warn_threshold"] == 30.0
    assert stats["block_threshold"] == 60.0


def test_threat_intel_signals_feed_correctly() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal(
        "s1",
        _sig(
            category="prompt_injection",
            confidence=0.9,
            severity="high",
            source="threat_intel",
            description="threat match",
        ),
    )
    out = acc.evaluate("s1")
    assert out.total_signals == 1
    assert out.contributing_signals[0]["source"] == "threat_intel"


def test_composite_attack_scenario() -> None:
    acc = SessionRiskAccumulator()
    categories = ["prompt_injection", "data_exfiltration", "path_traversal", "tool_abuse", "resource_abuse"]
    for cat in categories:
        acc.record_signal("s1", _sig(category=cat, severity="low", confidence=1.0))
    out = acc.evaluate("s1")
    assert out.action == "block"
    assert out.composite_score >= 60.0


def test_benign_session_stays_low() -> None:
    acc = SessionRiskAccumulator()
    for _ in range(20):
        out = acc.evaluate("benign")
        assert out.composite_score < 30.0
    out = acc.evaluate("benign")
    assert out.action == "allow"


def test_mixed_benign_and_suspicious() -> None:
    acc = SessionRiskAccumulator()
    for _ in range(3):
        _ = acc.evaluate("s1")
    acc.record_signal("s1", _sig(category="prompt_injection", severity="low", confidence=0.9))
    acc.record_signal("s1", _sig(category="data_exfiltration", severity="low", confidence=0.9))
    out = acc.evaluate("s1")
    assert out.action in {"allow", "warn", "block"}
    assert out.composite_score > 0.0


def test_stats_tracking() -> None:
    acc = SessionRiskAccumulator()
    acc.record_signal("s1", _sig())
    _ = acc.evaluate("s1")
    stats = acc.stats
    assert stats["signals_recorded"] == 1
    assert stats["total_evaluations"] >= 1
    assert stats["sessions_tracked"] == 1


def test_stats_escalation_counts() -> None:
    acc = SessionRiskAccumulator(warn_threshold=10.0, block_threshold=30.0)
    acc.record_signal("s1", _sig(category="a", severity="medium", confidence=1.0))
    acc.record_signal("s1", _sig(category="b", severity="high", confidence=1.0))
    stats = acc.stats
    assert stats["escalations_warn"] >= 1
    assert stats["escalations_block"] >= 1



"""Demo mode server for dashboard-only experience."""

from __future__ import annotations

import json
import random
import threading
import time
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Any
from urllib.parse import urlsplit

from orchesis.dashboard import get_dashboard_html


DEMO_STATS: dict[str, Any] = {
    "requests_total": 14847,
    "requests_blocked": 23,
    "requests_allowed": 14824,
    "tokens_saved": 2891443,
    "money_saved_usd": 43.37,
    "cache_hits": 1204,
    "cache_hit_rate": 0.081,
    "active_sessions": 7,
    "agents_discovered": 12,
    "fleet_health": "B+",
    "cost_velocity": {"current_rate_per_hour": 1.24, "projection_24h": 29.76, "is_anomalous": False},
    "threat_intel": {
        "signatures_loaded": 25,
        "blocked_by_category": {
            "prompt_injection": 8,
            "credential_leak": 5,
            "tool_abuse": 4,
            "privilege_escalation": 3,
            "data_exfiltration": 2,
            "loop_attack": 1,
        },
    },
    "top_models": [
        {"model": "gpt-4o", "requests": 8421, "cost": 12.63},
        {"model": "claude-sonnet-4-20250514", "requests": 3891, "cost": 7.78},
        {"model": "gpt-4o-mini", "requests": 2535, "cost": 0.51},
    ],
    "mast_compliance": {
        "covered": 11,
        "total": 14,
        "score": 78.6,
        "gaps": ["M12: Cascading Failure", "M13: Resource Exhaustion", "M14: Emergent Behavior"],
    },
    "owasp_coverage": {"covered": 8, "total": 10, "score": 80.0},
    "approvals_pending": 2,
    "uptime_seconds": 86400,
}


class DemoServer:
    """Serves dashboard with deterministic, realistic mock data."""

    def __init__(self) -> None:
        self._start = time.time()
        self._lock = threading.Lock()
        self._pending_approvals: dict[str, dict[str, Any]] = {
            "demo-approval-1": {
                "approval_id": "demo-approval-1",
                "timestamp": time.time() - 180,
                "agent_id": "deploy-bot",
                "tool_name": "system.run",
                "tool_args": {"action": "rm -rf /tmp/cache"},
                "reason": "Destructive command detected",
                "risk": "high",
            },
            "demo-approval-2": {
                "approval_id": "demo-approval-2",
                "timestamp": time.time() - 720,
                "agent_id": "data-sync",
                "tool_name": "api.call",
                "tool_args": {"domain": "internal.corp.com"},
                "reason": "Blocked domain match",
                "risk": "medium",
            },
        }
        self._approval_history: list[dict[str, Any]] = [
            {"status": "approved", "agent_id": "api-monitor", "tool_name": "api.call", "timestamp": time.time() - 7200},
            {"status": "denied", "agent_id": "refactor-agent", "tool_name": "system.run", "timestamp": time.time() - 18000},
        ]

    def _handle_stats(self) -> dict[str, Any]:
        jitter = random.Random(int(time.time() // 2))
        base_rate = float(DEMO_STATS["cost_velocity"]["current_rate_per_hour"])
        rate = max(0.1, base_rate + jitter.uniform(-0.08, 0.08))
        projection = rate * 24.0
        with self._lock:
            pending = list(self._pending_approvals.values())
            history = list(self._approval_history)
        return {
            "requests": int(DEMO_STATS["requests_total"] + jitter.randint(-4, 4)),
            "blocked": int(DEMO_STATS["requests_blocked"]),
            "cost_today": round(20.14 + jitter.uniform(-0.15, 0.15), 4),
            "cache_hit_rate_percent": round(float(DEMO_STATS["cache_hit_rate"]) * 100.0, 2),
            "semantic_cache": {
                "enabled": True,
                "hit_rate_percent": round(float(DEMO_STATS["cache_hit_rate"]) * 100.0, 2),
                "exact_hits": 889,
                "semantic_hits": 315,
                "misses": 13542,
                "entries": 202,
                "max_entries": 1000,
                "total_tokens_saved": int(DEMO_STATS["tokens_saved"]),
                "total_cost_saved_usd": float(DEMO_STATS["money_saved_usd"]),
            },
            "threat_intel": {
                "enabled": True,
                "total_signatures": int(DEMO_STATS["threat_intel"]["signatures_loaded"]),
                "total_scans": int(DEMO_STATS["requests_total"]),
                "total_matches": int(DEMO_STATS["requests_blocked"]),
                "blocks": int(DEMO_STATS["requests_blocked"]),
                "matches_by_category": dict(DEMO_STATS["threat_intel"]["blocked_by_category"]),
                "matches_by_severity": {"critical": 2, "high": 7, "medium": 11, "low": 3},
                "top_threats": [["Prompt Injection", 8], ["Credential Leak", 5], ["Tool Abuse", 4]],
            },
            "experiments": {"active": 2, "total_assignments": 643},
            "task_tracking": {"overall_success_rate": 0.94, "tracked_sessions": 123, "outcomes": {"success": 114, "failure": 5, "loop": 1, "abandoned": 2, "timeout": 1, "escalated": 0}},
            "context_engine": {"enabled": True, "total_tokens_saved": int(DEMO_STATS["tokens_saved"]), "strategy_hits": {"dedup_system_prompt": 321, "remove_stale_messages": 198}},
            "context_optimizer": {
                "enabled": True,
                "total_original_tokens": 5400000,
                "total_optimized_tokens": 2508557,
            },
            "cost_velocity": {
                "current_rate_per_hour": round(rate, 4),
                "projection_24h": round(projection, 4),
                "avg_7d_rate": 1.02,
                "is_anomalous": False,
            },
            "agent_discovery": {
                "enabled": True,
                "agents_tracked": int(DEMO_STATS["agents_discovered"]),
            },
            "tool_policy": {
                "approvals": {"pending_count": len(pending), "approved_count": 12, "denied_count": 3, "avg_wait_seconds": 211.0}
            },
            "compliance": {
                "frameworks": {
                    "owasp_llm_top10_2025": {"percent": 80.0},
                    "nist_ai_rmf_1_0": {"percent": 50.0},
                }
            },
            "demo_approvals_history": history,
        }

    def _overview(self) -> dict[str, Any]:
        stats = self._handle_stats()
        now = time.time()
        return {
            "status": "monitoring",
            "uptime_seconds": max(1.0, now - self._start),
            "total_requests": stats["requests"],
            "blocked_requests": stats["blocked"],
            "total_cost_usd": stats["cost_today"],
            "active_agents": int(DEMO_STATS["agents_discovered"]),
            "fleet_health": str(DEMO_STATS["fleet_health"]),
            "money_saved_usd": float(DEMO_STATS["money_saved_usd"]),
            "cost_velocity": dict(stats["cost_velocity"]),
            "compliance_overview": {
                "mast": {"score": 78.6, "covered": 11, "total": 14, "gaps": ["M12 Cascading Failure", "M13 Resource Exhaustion", "M14 Emergent Behavior"]},
                "owasp_agentic_ai": {"score": 80.0, "covered": 8, "total": 10, "gaps": ["ASI-09 Insufficient Logging", "ASI-10 Unsafe Plugin Design"]},
                "eu_ai_act": {"score": 60.0, "audit_trail": True, "incident_reporting": True, "risk_assessment": "partial", "human_oversight": "partial", "documentation": False},
                "nist_ai_rmf": {"score": 50.0, "govern": "full", "map": "partial", "measure": "full", "manage": "partial"},
            },
            "circuit_breakers": {"default": {"state": "closed", "failures": 0}},
            "budget": {"limit_usd": 10.0, "spent_usd": stats["cost_today"], "remaining_usd": max(0.0, 10.0 - float(stats["cost_today"]))},
            "recent_events": [
                {"timestamp": now - 50, "type": "blocked", "severity": "high", "description": "Blocked prompt injection attempt"},
                {"timestamp": now - 180, "type": "warn", "severity": "medium", "description": "Tool approval required for system.run"},
            ],
            "cost_timeline": [{"timestamp": now - 3600 + i * 300, "cumulative_cost": round(0.8 + i * 0.13, 4)} for i in range(12)],
            "flow_xray": {"enabled": True},
            "connection_pool": {"active": 4, "total_connections": 12, "hits": 1204, "misses": 275, "pools": {"api.openai.com": 7}},
            "savings": {"cache_savings": 20.5, "cascade_savings": 11.2, "loop_savings": 2.1, "total_savings": 33.8},
        }

    def _json(self, handler: BaseHTTPRequestHandler, status: int, payload: dict[str, Any]) -> None:
        raw = json.dumps(payload).encode("utf-8")
        handler.send_response(status)
        handler.send_header("Content-Type", "application/json")
        handler.send_header("Content-Length", str(len(raw)))
        handler.end_headers()
        handler.wfile.write(raw)

    def start(self, port: int = 8080) -> None:
        server = self

        class _Handler(BaseHTTPRequestHandler):
            def do_GET(self) -> None:  # noqa: N802
                parsed = urlsplit(self.path)
                path = parsed.path
                if path in {"/", "/dashboard", "/dashboard/"}:
                    html = get_dashboard_html(demo_mode=True).encode("utf-8")
                    self.send_response(200)
                    self.send_header("Content-Type", "text/html; charset=utf-8")
                    self.send_header("Content-Length", str(len(html)))
                    self.end_headers()
                    self.wfile.write(html)
                    return
                if path == "/health":
                    server._json(self, 200, {"status": "ok", "mode": "demo"})
                    return
                if path in {"/stats", "/api/v1/stats"}:
                    server._json(self, 200, server._handle_stats())
                    return
                if path == "/api/dashboard/overview":
                    server._json(self, 200, server._overview())
                    return
                if path == "/api/dashboard/agents":
                    server._json(
                        self,
                        200,
                        {"agents": [{"agent_id": "deploy-bot", "state": "active", "total_requests": 382, "avg_tokens": 914, "anomaly_score": 0.08, "tools_used": ["system.run", "web_fetch"], "last_seen": "just now"}]},
                    )
                    return
                if path in {"/api/sessions", "/sessions"}:
                    server._json(self, 200, {"sessions": []})
                    return
                if path == "/api/flow/sessions":
                    server._json(self, 200, {"sessions": [{"id": "demo-session-1"}]})
                    return
                if path.startswith("/api/flow/analyze/"):
                    server._json(self, 200, {"topology": {"depth": 3, "width": 2, "density": 0.44, "total_cost_usd": 1.83, "critical_path": ["node-a", "node-b"]}, "summary": {"health_score": 0.82}, "patterns": []})
                    return
                if path.startswith("/api/flow/graph/"):
                    server._json(self, 200, {"nodes": [{"node_id": "node-a", "node_type": "llm", "model": "gpt-4o", "cost_usd": 0.12}], "edges": []})
                    return
                if path == "/api/experiments":
                    server._json(self, 200, {"experiments": [{"experiment_id": "exp-1", "name": "Routing AB", "status": "running"}]})
                    return
                if path == "/api/tasks/outcomes":
                    server._json(self, 200, {"ok": True})
                    return
                if path == "/api/tasks/correlations":
                    server._json(self, 200, {"insights": ["Tool-approved flows have lower failure rate in demo data."]})
                    return
                if path == "/api/threats":
                    server._json(self, 200, {"threats": [{"threat_id": "ORCH-TA-001", "name": "Prompt Injection", "category": "prompt_injection", "severity": "high"}]})
                    return
                if path == "/api/threats/stats":
                    stats = server._handle_stats()["threat_intel"]
                    server._json(self, 200, stats)
                    return
                if path == "/api/compliance/summary":
                    server._json(self, 200, {"frameworks": {"owasp_llm_top10_2025": {"percent": 80.0}, "nist_ai_rmf_1_0": {"percent": 50.0}}})
                    return
                if path == "/api/compliance/coverage":
                    server._json(self, 200, {"frameworks": {"owasp_llm_top10_2025": {"items": [{"id": "ASI-01", "name": "Authorization", "status": "covered", "modules": ["policy_engine"]}]}}})
                    return
                if path.startswith("/api/compliance/findings"):
                    server._json(self, 200, {"findings": [{"timestamp": "now", "severity": "medium", "description": "Demo gap: ASI-10 unsafe plugin design", "framework_mappings": [["owasp_llm_top10_2025", "ASI-10"]]}]})
                    return
                if path.startswith("/api/compliance/report"):
                    server._json(self, 200, {"report": "demo"})
                    return
                if path == "/api/v1/savings":
                    server._json(self, 200, {"cache_savings": 20.5, "cascade_savings": 11.2, "loop_savings": 2.1, "total_savings": 33.8})
                    return
                if path == "/api/v1/approvals":
                    with server._lock:
                        pending = list(server._pending_approvals.values())
                        history = list(server._approval_history)
                    server._json(self, 200, {"pending": pending, "history": history, "stats": {"pending_count": len(pending)}})
                    return
                server._json(self, 404, {"error": "not_found"})

            def do_POST(self) -> None:  # noqa: N802
                parsed = urlsplit(self.path)
                path = parsed.path
                if path.startswith("/api/v1/approvals/") and path.endswith("/approve"):
                    request_id = path.split("/api/v1/approvals/", 1)[1].rsplit("/approve", 1)[0].strip("/")
                    with server._lock:
                        item = server._pending_approvals.pop(request_id, None)
                        ok = item is not None
                        if ok:
                            server._approval_history.insert(0, {"status": "approved", "agent_id": item["agent_id"], "tool_name": item["tool_name"], "timestamp": time.time()})
                    server._json(self, 200 if ok else 404, {"approved": ok, "approval_id": request_id})
                    return
                if path.startswith("/api/v1/approvals/") and path.endswith("/deny"):
                    request_id = path.split("/api/v1/approvals/", 1)[1].rsplit("/deny", 1)[0].strip("/")
                    with server._lock:
                        item = server._pending_approvals.pop(request_id, None)
                        ok = item is not None
                        if ok:
                            server._approval_history.insert(0, {"status": "denied", "agent_id": item["agent_id"], "tool_name": item["tool_name"], "timestamp": time.time()})
                    server._json(self, 200 if ok else 404, {"denied": ok, "approval_id": request_id})
                    return
                server._json(self, 404, {"error": "not_found"})

            def log_message(self, fmt: str, *args: Any) -> None:
                _ = (fmt, args)

        httpd = ThreadingHTTPServer(("0.0.0.0", max(1, int(port))), _Handler)
        print(f"Orchesis demo dashboard running at http://127.0.0.1:{int(port)}/dashboard")
        httpd.serve_forever()




from __future__ import annotations

import json
import os
from pathlib import Path
import socket
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.error import HTTPError
from urllib.request import Request as UrlRequest, urlopen

import pytest

from orchesis.dashboard import get_dashboard_html
from orchesis.proxy import HTTPProxyConfig, LLMHTTPProxy


class _DashboardUpstreamHandler(BaseHTTPRequestHandler):
    def do_POST(self) -> None:  # noqa: N802
        length = int(self.headers.get("Content-Length", "0") or "0")
        _ = self.rfile.read(length)
        payload = {
            "model": "gpt-4o-mini",
            "usage": {"prompt_tokens": 6, "completion_tokens": 4},
            "choices": [{"message": {"content": "ok"}, "finish_reason": "stop"}],
        }
        data = json.dumps(payload).encode("utf-8")
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.wfile.write(data)

    def log_message(self, fmt: str, *args) -> None:
        _ = (fmt, args)


def _start_http_server(handler_cls: type[BaseHTTPRequestHandler]) -> tuple[HTTPServer, threading.Thread]:
    server = HTTPServer(("127.0.0.1", 0), handler_cls)
    thread = threading.Thread(target=server.serve_forever, daemon=True)
    thread.start()
    return server, thread


def _pick_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    p = int(sock.getsockname()[1])
    sock.close()
    return p


def _make_proxy(tmp_path: Path, policy_text: str) -> tuple[LLMHTTPProxy, HTTPServer]:
    upstream, _ = _start_http_server(_DashboardUpstreamHandler)
    policy = tmp_path / "policy.yaml"
    policy.write_text(policy_text, encoding="utf-8")
    port = _pick_port()
    proxy = LLMHTTPProxy(
        policy_path=str(policy),
        config=HTTPProxyConfig(
            host="127.0.0.1",
            port=port,
            upstream={
                "openai": f"http://127.0.0.1:{upstream.server_address[1]}",
                "anthropic": f"http://127.0.0.1:{upstream.server_address[1]}",
            },
        ),
    )
    proxy.start(blocking=False)
    return proxy, upstream


def _get_json(port: int, path: str) -> tuple[int, dict]:
    with urlopen(f"http://127.0.0.1:{port}{path}", timeout=5) as resp:
        return int(resp.status), json.loads(resp.read().decode("utf-8"))


def _get_text(port: int, path: str) -> tuple[int, str, str]:
    with urlopen(f"http://127.0.0.1:{port}{path}", timeout=5) as resp:
        body = resp.read().decode("utf-8")
        ctype = str(resp.headers.get("Content-Type", ""))
        return int(resp.status), ctype, body


def _post_chat(port: int, *, agent_id: str | None = None) -> None:
    headers = {"Content-Type": "application/json", "Authorization": "Bearer x"}
    if agent_id:
        headers["X-Agent-Id"] = agent_id
    req = UrlRequest(
        f"http://127.0.0.1:{port}/v1/chat/completions",
        data=json.dumps({"model": "gpt-4o", "messages": [{"role": "user", "content": "hello"}]}).encode("utf-8"),
        headers=headers,
        method="POST",
    )
    with urlopen(req, timeout=5) as resp:
        _ = resp.read()


# HTML generation tests (5)
def test_dashboard_html_returns_valid_html_string() -> None:
    html = get_dashboard_html()
    assert isinstance(html, str)
    assert "<html" in html.lower()
    assert "</html>" in html.lower()


def test_dashboard_html_contains_expected_tabs() -> None:
    html = get_dashboard_html().lower()
    assert "shield" in html
    assert "agents" in html
    assert "sessions" in html
    assert "flow x-ray" in html


def test_dashboard_html_contains_css_styles() -> None:
    html = get_dashboard_html()
    assert "<style>" in html
    assert "#0A0A12" in html
    assert "backdrop-filter" in html


def test_dashboard_html_contains_javascript() -> None:
    html = get_dashboard_html()
    assert "<script>" in html
    assert "POLL_INTERVAL" in html
    assert "fetchData(" in html
    assert html.count("const ratio =") == 1
    assert "const poolHitRatio =" in html


def test_dashboard_html_size_reasonable() -> None:
    html = get_dashboard_html()
    assert len(html) > 5_000
    assert len(html) < 500_000


# Dashboard v2 tests (10)
def test_dashboard_html_contains_experiments_tab() -> None:
    html = get_dashboard_html()
    assert "Experiments" in html or "experiments" in html


def test_dashboard_html_contains_threats_tab() -> None:
    html = get_dashboard_html()
    assert "Threats" in html or "threats" in html


def test_dashboard_html_contains_cache_tab() -> None:
    html = get_dashboard_html()
    assert "Cache" in html or "cache" in html


def test_dashboard_experiments_section_ids() -> None:
    html = get_dashboard_html()
    assert "exp-active" in html
    assert "exp-cards" in html
    assert "exp-correlations" in html


def test_dashboard_threats_section_ids() -> None:
    html = get_dashboard_html()
    assert "th-sigs" in html
    assert "th-scans" in html
    assert "th-matches" in html
    assert "th-blocks" in html


def test_dashboard_cache_section_ids() -> None:
    html = get_dashboard_html()
    assert "c-hit-rate" in html
    assert "c-tokens" in html
    assert "c-cost" in html
    assert "c-entries" in html


def test_dashboard_shield_new_metrics() -> None:
    html = get_dashboard_html()
    assert "m-threats" in html
    assert "m-cache-rate" in html
    assert "m-experiments" in html
    assert "m-task-success" in html


def test_dashboard_poll_functions_exist() -> None:
    html = get_dashboard_html()
    assert "pollExperiments" in html
    assert "pollThreats" in html
    assert "pollCache" in html


def test_dashboard_severity_css_classes() -> None:
    html = get_dashboard_html()
    assert "sev-critical" in html
    assert "sev-high" in html
    assert "sev-medium" in html
    assert "sev-low" in html


def test_dashboard_variant_card_css() -> None:
    html = get_dashboard_html()
    assert "variant-card" in html
    assert "variant-pair" in html


def test_compliance_tab_renders() -> None:
    html = get_dashboard_html()
    assert "compliance" in html.lower()
    assert "cmp-overview-text" in html


def test_approvals_tab_renders() -> None:
    html = get_dashboard_html()
    assert "approvals" in html.lower()
    assert "approvals-pending" in html


def test_approvals_approve_button() -> None:
    html = get_dashboard_html()
    assert "APPROVE" in html


def test_approvals_deny_button() -> None:
    html = get_dashboard_html()
    assert "DENY" in html


# Overview endpoint tests (8)
def test_dashboard_overview_returns_expected_keys(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        code, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert code == 200
        for key in (
            "status",
            "uptime_seconds",
            "total_requests",
            "blocked_requests",
            "total_cost_usd",
            "active_agents",
            "circuit_breakers",
            "recent_events",
            "cost_timeline",
        ):
            assert key in payload
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_status_clear_without_blocks(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["status"] == "clear"
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_status_monitoring_with_recent_blocks(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        proxy._add_dashboard_event("blocked", "medium", "blocked test")
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["status"] == "monitoring"
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_status_alert_with_circuit_open(tmp_path: Path) -> None:
    policy = "rules: []\ncircuit_breaker:\n  enabled: true\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        with proxy._circuit_breaker._lock:  # noqa: SLF001
            proxy._circuit_breaker._state = proxy._circuit_breaker.STATE_OPEN  # noqa: SLF001
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["status"] == "alert"
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_cost_timeline_list(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert isinstance(payload["cost_timeline"], list)
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_recent_events_list(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert isinstance(payload["recent_events"], list)
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_budget_present_when_configured(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\nbudgets:\n  daily: 10\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["budget"]["limit_usd"] == 10.0
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_overview_budget_zero_when_not_configured(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["budget"]["limit_usd"] == 0.0
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


# Agents endpoint tests (5)
def test_dashboard_agents_returns_agents_key(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        assert "agents" in payload
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_agents_empty_when_detector_disabled(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        assert payload["agents"] == []
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_agents_contains_expected_fields(tmp_path: Path) -> None:
    policy = "rules: []\nbehavioral_fingerprint:\n  enabled: true\n  learning_window: 1\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port, agent_id="agent-a")
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        assert payload["agents"]
        agent = payload["agents"][0]
        for key in ("agent_id", "total_requests", "avg_tokens", "anomaly_score", "tools_used", "last_seen"):
            assert key in agent
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_agents_multiple_agents(tmp_path: Path) -> None:
    policy = "rules: []\nbehavioral_fingerprint:\n  enabled: true\n  learning_window: 1\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port, agent_id="agent-a")
        _post_chat(proxy._config.port, agent_id="agent-b")
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        ids = {item["agent_id"] for item in payload["agents"]}
        assert "agent-a" in ids and "agent-b" in ids
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_agents_anomaly_scores_present(tmp_path: Path) -> None:
    policy = "rules: []\nbehavioral_fingerprint:\n  enabled: true\n  learning_window: 1\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port, agent_id="agent-a")
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        assert "anomaly_score" in payload["agents"][0]
        assert "anomaly_scores" in payload["agents"][0]
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


# Proxy route handling (5)
def test_dashboard_route_returns_html(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        code, ctype, _ = _get_text(proxy._config.port, "/dashboard")
        assert code == 200
        assert "text/html" in ctype
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_trailing_slash_returns_html(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        code, ctype, _ = _get_text(proxy._config.port, "/dashboard/")
        assert code == 200
        assert "text/html" in ctype
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_html_response_non_empty(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _, _, body = _get_text(proxy._config.port, "/dashboard")
        assert len(body) > 0
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_unknown_subpath_returns_404(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        with pytest.raises(HTTPError) as err:
            _get_text(proxy._config.port, "/dashboard/xxx")
        assert err.value.code == 404
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_dashboard_accessible_without_sessions(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        code, _, _ = _get_text(proxy._config.port, "/dashboard")
        assert code == 200
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_favicon_path_returns_no_content(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        req = UrlRequest(f"http://127.0.0.1:{proxy._config.port}/favicon.ico")
        with urlopen(req, timeout=5) as resp:
            assert int(resp.status) == 204
            assert resp.read() == b""
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


# Integration (5)
def test_overview_reflects_proxy_stats_after_requests(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        _post_chat(proxy._config.port)
        _post_chat(proxy._config.port)
        _, payload = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert payload["total_requests"] >= 2
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_agent_list_updates_after_behavioral_recording(tmp_path: Path) -> None:
    policy = "rules: []\nbehavioral_fingerprint:\n  enabled: true\n  learning_window: 1\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port, agent_id="agent-z")
        _, payload = _get_json(proxy._config.port, "/api/dashboard/agents")
        assert any(item["agent_id"] == "agent-z" for item in payload["agents"])
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_flow_xray_data_accessible_from_dashboard_paths(tmp_path: Path) -> None:
    policy = "rules: []\nflow_xray:\n  enabled: true\n"
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port)
        _, overview = _get_json(proxy._config.port, "/api/dashboard/overview")
        assert "flow_xray" in overview
        _, flow_sessions = _get_json(proxy._config.port, "/api/flow/sessions")
        assert "sessions" in flow_sessions
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()


def test_sessions_list_consistent_between_api_paths(tmp_path: Path) -> None:
    policy = "rules: []\nrecording:\n  enabled: true\n"
    cwd = Path.cwd()
    os.chdir(tmp_path)
    proxy, upstream = _make_proxy(tmp_path, policy)
    try:
        _post_chat(proxy._config.port)
        _, a = _get_json(proxy._config.port, "/sessions")
        _, b = _get_json(proxy._config.port, "/api/sessions")
        assert len(a.get("sessions", [])) == len(b.get("sessions", []))
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()
        os.chdir(cwd)


def test_stats_endpoint_still_works(tmp_path: Path) -> None:
    proxy, upstream = _make_proxy(tmp_path, "rules: []\n")
    try:
        code, payload = _get_json(proxy._config.port, "/stats")
        assert code == 200
        assert "requests" in payload
    finally:
        proxy.stop()
        upstream.shutdown()
        upstream.server_close()



"""Structured JSON logging for Orchesis components."""

from __future__ import annotations

import json
import sys
from datetime import datetime, timezone
from typing import Any, TextIO

_LEVEL_ORDER = {"DEBUG": 10, "INFO": 20, "WARN": 30, "ERROR": 40}


class StructuredLogger:
    """JSON-structured logger for Orchesis components."""

    def __init__(self, component: str, output: str = "stderr", level: str = "INFO"):
        self._component = component
        self._level = level.strip().upper() if isinstance(level, str) else "INFO"
        self._stream: TextIO = sys.stderr if output != "stdout" else sys.stdout

    def info(self, message: str, **context) -> None:
        self._emit("INFO", message, **context)

    def warn(self, message: str, **context) -> None:
        self._emit("WARN", message, **context)

    def error(self, message: str, **context) -> None:
        self._emit("ERROR", message, **context)

    def debug(self, message: str, **context) -> None:
        self._emit("DEBUG", message, **context)

    def _emit(self, level: str, message: str, **context) -> None:
        configured = _LEVEL_ORDER.get(self._level, _LEVEL_ORDER["INFO"])
        current = _LEVEL_ORDER.get(level, _LEVEL_ORDER["INFO"])
        if current < configured:
            return
        payload: dict[str, Any] = {
            "ts": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            "level": level,
            "component": self._component,
            "msg": message,
        }
        payload.update(context)
        self._stream.write(json.dumps(payload, ensure_ascii=False) + "\n")
        self._stream.flush()



"""Adaptive model cascade routing with response cache."""

from __future__ import annotations

from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from datetime import datetime
import hashlib
import json
import threading
import time
from enum import IntEnum
from typing import Any

from orchesis.request_parser import ParsedRequest, ParsedResponse


class CascadeLevel(IntEnum):
    TRIVIAL = 0
    SIMPLE = 1
    MEDIUM = 2
    COMPLEX = 3


_LEVEL_NAME_TO_ENUM = {
    "trivial": CascadeLevel.TRIVIAL,
    "simple": CascadeLevel.SIMPLE,
    "medium": CascadeLevel.MEDIUM,
    "complex": CascadeLevel.COMPLEX,
}
_LEVEL_ENUM_TO_NAME = {value: key for key, value in _LEVEL_NAME_TO_ENUM.items()}


@dataclass
class CascadeDecision:
    model: str
    max_tokens: int
    cascade_level: CascadeLevel
    from_cache: bool = False
    cache_key: str = ""


class ResponseCache:
    """Simple in-memory LRU cache with TTL."""

    def __init__(self, ttl_seconds: int = 300, max_entries: int = 1000) -> None:
        self._ttl_seconds = max(1, int(ttl_seconds))
        self._max_entries = max(1, int(max_entries))
        self._lock = threading.Lock()
        self._items: OrderedDict[str, tuple[float, bytes]] = OrderedDict()

    def get(self, key: str) -> bytes | None:
        now = time.time()
        with self._lock:
            item = self._items.get(key)
            if item is None:
                return None
            ts, payload = item
            if (now - ts) > self._ttl_seconds:
                self._items.pop(key, None)
                return None
            self._items.move_to_end(key)
            return payload

    def set(self, key: str, value: bytes) -> None:
        with self._lock:
            self._items[key] = (time.time(), value)
            self._items.move_to_end(key)
            while len(self._items) > self._max_entries:
                self._items.popitem(last=False)

    def size(self) -> int:
        with self._lock:
            return len(self._items)

    def evict_expired(self) -> int:
        now = time.time()
        with self._lock:
            expired = [k for k, (ts, _) in self._items.items() if (now - ts) > self._ttl_seconds]
            for key in expired:
                self._items.pop(key, None)
            return len(expired)


class CascadeClassifier:
    """Classify request complexity using stdlib heuristics."""

    _KEYWORDS = {
        "code": CascadeLevel.MEDIUM,
        "python": CascadeLevel.MEDIUM,
        "javascript": CascadeLevel.MEDIUM,
        "typescript": CascadeLevel.MEDIUM,
        "math": CascadeLevel.MEDIUM,
        "equation": CascadeLevel.MEDIUM,
        "analyze": CascadeLevel.MEDIUM,
        "compare": CascadeLevel.MEDIUM,
        "research": CascadeLevel.MEDIUM,
        "debug": CascadeLevel.MEDIUM,
        "refactor": CascadeLevel.COMPLEX,
        "multi-step": CascadeLevel.COMPLEX,
    }

    def __init__(self) -> None:
        self._task_history: OrderedDict[str, str] = OrderedDict()
        self._max_history = 10000
        self._lock = threading.Lock()

    @staticmethod
    def _approx_tokens(text: str) -> int:
        if not isinstance(text, str) or not text:
            return 0
        return max(0, int(len(text) / 4))

    def classify(self, parsed_request: ParsedRequest, context: dict[str, Any] | None = None) -> CascadeLevel:
        safe_context = context if isinstance(context, dict) else {}
        text = parsed_request.content_text if isinstance(parsed_request.content_text, str) else ""
        token_estimate = self._approx_tokens(text)
        tool_calls_count = len(parsed_request.tool_calls)
        tool_definitions_count = len(parsed_request.tool_definitions)

        if token_estimate == 0 and tool_calls_count == 0:
            return CascadeLevel.TRIVIAL

        level = CascadeLevel.SIMPLE

        if token_estimate > 2000:
            level = max(level, CascadeLevel.COMPLEX)
        elif token_estimate >= 500:
            level = max(level, CascadeLevel.MEDIUM)

        if tool_calls_count >= 3:
            level = max(level, CascadeLevel.COMPLEX)
        elif tool_calls_count >= 1:
            level = max(level, CascadeLevel.MEDIUM)

        if tool_definitions_count >= 20:
            level = max(level, CascadeLevel.MEDIUM)

        lowered = text.lower()
        for keyword, target_level in self._KEYWORDS.items():
            if keyword in lowered:
                level = max(level, target_level)

        task_id = safe_context.get("task_id")
        if isinstance(task_id, str) and task_id.strip():
            content_hash = hashlib.sha256(lowered.encode("utf-8")).hexdigest()
            with self._lock:
                previous_hash = self._task_history.get(task_id)
                self._task_history[task_id] = content_hash
                self._task_history.move_to_end(task_id)
                while len(self._task_history) > self._max_history:
                    self._task_history.popitem(last=False)
            if isinstance(previous_hash, str) and previous_hash and previous_hash != content_hash:
                level = min(CascadeLevel.COMPLEX, CascadeLevel(level + 1))

        if bool(safe_context.get("previous_failed_attempts", False)):
            level = min(CascadeLevel.COMPLEX, CascadeLevel(level + 1))

        return level


class CascadeRouter:
    """Route requests to optimal model based on cascade level."""

    def __init__(self, config: dict[str, Any] | None = None) -> None:
        safe_cfg = config if isinstance(config, dict) else {}
        self._enabled = bool(safe_cfg.get("enabled", False))
        self._levels = safe_cfg.get("levels") if isinstance(safe_cfg.get("levels"), dict) else {}
        auto = safe_cfg.get("auto_escalate")
        self._auto_escalate = auto if isinstance(auto, dict) else {}
        cache_cfg = safe_cfg.get("cache")
        self._cache_cfg = cache_cfg if isinstance(cache_cfg, dict) else {}

        self._classifier = CascadeClassifier()
        self._cache = ResponseCache(
            ttl_seconds=int(self._cache_cfg.get("ttl_seconds", 300)),
            max_entries=int(self._cache_cfg.get("max_entries", 1000)),
        )
        self._cache_enabled = bool(self._cache_cfg.get("enabled", True))
        self._requests_by_level: dict[str, int] = defaultdict(int)
        self._cache_hits_by_level: dict[str, int] = defaultdict(int)
        self._cache_lookups = 0
        self._cache_hits = 0
        self._lock = threading.Lock()

    @property
    def enabled(self) -> bool:
        return self._enabled

    def _level_name(self, level: CascadeLevel) -> str:
        return _LEVEL_ENUM_TO_NAME.get(level, "simple")

    def _level_cfg(self, level: CascadeLevel) -> dict[str, Any]:
        return self._levels.get(self._level_name(level), {}) if isinstance(self._levels, dict) else {}

    def make_cache_key(self, parsed_request: ParsedRequest, model: str) -> str:
        msg = {
            "model": model,
            "messages": parsed_request.messages,
            "tool_definitions": parsed_request.tool_definitions,
        }
        raw = json.dumps(msg, ensure_ascii=False, sort_keys=True, separators=(",", ":"))
        return hashlib.sha256(raw.encode("utf-8")).hexdigest()

    def get_cache(self, key: str, level: CascadeLevel) -> bytes | None:
        if not self._enabled or not self._cache_enabled:
            return None
        with self._lock:
            self._cache_lookups += 1
        payload = self._cache.get(key)
        if payload is not None:
            with self._lock:
                self._cache_hits += 1
                self._cache_hits_by_level[self._level_name(level)] += 1
        return payload

    def route(self, parsed_request: ParsedRequest, context: dict[str, Any] | None = None) -> CascadeDecision:
        if not self._enabled:
            model = parsed_request.model or "gpt-4o"
            max_tokens = int(parsed_request.raw_body.get("max_tokens", 0) or 0)
            return CascadeDecision(model=model, max_tokens=max_tokens, cascade_level=CascadeLevel.SIMPLE)

        level = self._classifier.classify(parsed_request, context=context)
        level_cfg = self._level_cfg(level)

        model = str(level_cfg.get("model", parsed_request.model or "gpt-4o"))
        max_tokens = int(level_cfg.get("max_tokens", parsed_request.raw_body.get("max_tokens", 0) or 0))
        if model == "" and parsed_request.model:
            model = parsed_request.model

        with self._lock:
            self._requests_by_level[self._level_name(level)] += 1

        return CascadeDecision(
            model=model,
            max_tokens=max_tokens,
            cascade_level=level,
            from_cache=False,
            cache_key=self.make_cache_key(parsed_request, model),
        )

    def classify(self, parsed_request: ParsedRequest, context: dict[str, Any] | None = None) -> CascadeLevel:
        return self._classifier.classify(parsed_request, context=context)

    def level_name(self, level: CascadeLevel) -> str:
        return self._level_name(level)

    def should_escalate(self, status_code: int, parsed_response: ParsedResponse | None = None) -> bool:
        if not self._enabled:
            return False
        if not bool(self._auto_escalate.get("enabled", False)):
            return False
        if bool(self._auto_escalate.get("on_error", True)) and status_code >= 400:
            return True
        if bool(self._auto_escalate.get("on_low_confidence", True)) and isinstance(parsed_response, ParsedResponse):
            stop_reason = (parsed_response.stop_reason or "").lower()
            if stop_reason in {"length", "max_tokens", "insufficient_confidence"}:
                return True
        return False

    def escalate(self, decision: CascadeDecision) -> CascadeDecision:
        next_level = min(CascadeLevel.COMPLEX, CascadeLevel(decision.cascade_level + 1))
        next_cfg = self._level_cfg(next_level)
        next_model = str(next_cfg.get("model", decision.model))
        next_max_tokens = int(next_cfg.get("max_tokens", decision.max_tokens))
        return CascadeDecision(
            model=next_model,
            max_tokens=next_max_tokens,
            cascade_level=next_level,
            from_cache=False,
            cache_key=decision.cache_key,
        )

    def cache_response(self, decision: CascadeDecision, response_body: bytes) -> None:
        if not self._enabled or not self._cache_enabled:
            return
        if decision.cascade_level not in {CascadeLevel.TRIVIAL, CascadeLevel.SIMPLE}:
            return
        if decision.cache_key:
            self._cache.set(decision.cache_key, response_body)

    def record_result(self, decision: CascadeDecision, parsed_response: ParsedResponse) -> None:
        _ = (decision, parsed_response)
        # Reserved for future adaptive tuning hooks.

    def get_stats(self) -> dict[str, Any]:
        if self._cache_enabled:
            self._cache.evict_expired()
        with self._lock:
            lookups = self._cache_lookups
            hits = self._cache_hits
            hit_rate = (hits / lookups * 100.0) if lookups > 0 else 0.0
            return {
                "requests_by_level": dict(self._requests_by_level),
                "hits_by_level": dict(self._cache_hits_by_level),
                "cache_hit_rate_percent": round(hit_rate, 2),
                "cache_entries_count": self._cache.size(),
            }
